<?php

namespace pareviewsh;

use \JsonSerializable;

/**
 * Base class for PAReview checks.
 *
 * This class implements common features for PAReview.sh checks.
 */
abstract class CheckBase implements CheckInterface, IssueStatisticsInterface, JsonSerializable {

  /**
   * Issues found while conducting this check.
   *
   * @var \pareviewsh\CheckIssue[];
   */
  protected $issues;

  /**
   * Whether the check was executed.
   *
   * @var int
   */
  protected $status;

  /**
   * Amount of notices.
   *
   * @var int
   */
  protected $notices;

  /**
   * Amount of warnings.
   *
   * @var int
   */
  protected $warnings;

  /**
   * Amount of errors.
   *
   * @var int
   */
  protected $errors;

  /**
   * {@inheritdoc}
   */
  public static function getPriority() {
    // Set a default priority of 0 for all checks.
    return 0;
  }

  /**
   * {@inheritdoc}
   */
  public static function requiresGit() {
    // Most checks don't require Git. So by default we set this
    // requirement to FALSE.
    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function __construct() {
    $this->restoreDefaults();
  }

  /**
   * {@inheritdoc}
   */
  public function run() {
    $status = self::CHECK_STATUS_SUCCESS;
    if ($this->errors > 0) {
      $status = self::CHECK_STATUS_ERRORS;
    }
    else if ($this->warnings > 0) {
      $status = self::CHECK_STATUS_WARNINGS;
    }
    $this->status = $status;
    return $this->status;
  }

  /**
   * {@inheritdoc}
   */
  public function getStatus() {
    return $this->status;
  }

  /**
   * {@inheritdoc}
   */
  public function addIssue(CheckIssue $issue) {
    $this->issues[] = $issue;

    // Whether this issue has no sub issues.
    if (!$issue->hasFileIssues()) {
      switch ($issue->getSeverity()) {
        case IssueInterface::SEVERITY_NOTICE:
          $this->notices++;
          break;
        case IssueInterface::SEVERITY_WARNING:
          $this->warnings++;
          break;
        case IssueInterface::SEVERITY_ERROR:
          $this->errors++;
          break;
      }
    }
    // The issue is a wrapper for sub issues.
    else {
      // Add the amount of subissue severities to
      // our internal severity counters.
      $this->notices += $issue->getNoticesCount();
      $this->warnings += $issue->getWarningsCount();
      $this->errors += $issue->getErrorsCount();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function hasIssues($run = TRUE) {
    if ($this->status == self::CHECK_STATUS_NO_RUN && $run) {
      $this->run();
    }

    return count($this->issues) > 0;
  }

  /**
   * {@inheritdoc}
   */
  public function getIssues() {
    return $this->issues;
  }

  /**
   * {@inheritdoc}
   */
  public function getNoticesCount() {
    return $this->notices;
  }

  /**
   * {@inheritdoc}
   */
  public function getWarningsCount() {
    return $this->warnings;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrorsCount() {
    return $this->errors;
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return [
      'has_issues' => $this->hasIssues(),
      'notices' => $this->notices,
      'warnings' => $this->warnings,
      'errors' => $this->errors,
      'issues' => $this->issues,
    ];
  }

  /**
   * Set issues.
   *
   * @param \pareviewsh\Issues $issues
   *   Issues instance.
   */
  protected function setIssues(Issues $issues) {
    $this->issues = $issues;
  }

  /**
   * Reset all properties to their initial values.
   */
  protected function restoreDefaults() {
    $this->notices = 0;
    $this->warnings = 0;
    $this->errors = 0;
    $this->issues = [];
    $this->status = CheckInterface::CHECK_STATUS_NO_RUN;
  }

}
