<?php

namespace pareviewsh;

/**
 * Interface for checks conducted by PAReview.sh.
 *
 * A check conducts several inspections and collects issues found.
 *
 * PAReview.sh will load all classes that implement this interface
 * and are registered within Composer's autoload.
 */
interface CheckInterface {

  /**
   * The check did not run yet.
   */
  const CHECK_STATUS_NO_RUN = -1;
  /**
   * Check was completed without any issues.
   */
  const CHECK_STATUS_SUCCESS = 0;
  /**
   * The check was completed with warnings.
   */
  const CHECK_STATUS_WARNINGS = 1;
  /**
   * The check was completed with errors.
   */
  const CHECK_STATUS_ERRORS = 2;
  /**
   * An unspecified error occurred while running the check.
   */
  const CHECK_STATUS_UNKNOWN = 3;

  /**
   * Whether this check should run only when checking a Git Repository.
   *
   * @return bool
   *   TRUE, if a Git repository is required, FALSE otherwise.
   */
  public static function requiresGit();

  /**
   * Return an unique check identifier.
   *
   * Multiple checks can be grouped by using `:` as separator. This
   * allows running checks of a certain group only.
   *
   * @return string
   *   Check identifier.
   */
  public static function getId();

  /**
   * Return a human-readable check name.
   *
   * @return string
   *   Check name.
   */
  public static function getName();

  /**
   * Return the check description.
   *
   * This description will be used to give a summary of what the check
   * does.
   *
   * @return string
   *   Check description.
   */
  public static function getDescription();

  /**
   * Return the priority of the check.
   *
   * @return int
   *   Priority.
   */
  public static function getPriority();

  /**
   * Constructs a check instance.
   */
  public function __construct();

  /**
   * Return the execution status of the check.
   *
   * @return int
   *   One of the status values as defined by the `CHECK_STATUS`
   *   constants.
   */
  public function getStatus();

  /**
   * Run the check.
   *
   * @return int
   *   One of the status values as defined by the `CHECK_STATUS`
   *   constants.
   */
  public function run();

  /**
   * Add an issue to the result set of this check.
   *
   * @param \pareviewsh\CheckIssue $issue
   *   Issue to add.
   *
   * @return bool
   *   TRUE, if issues where found, FALSE otherwise.
   */
  public function addIssue(CheckIssue $issue);

  /**
   * Whether this check produced issues.
   *
   * @param bool $run
   *   [Optional] Whether to run the check first, if it was not executed
   *   before. Defaults to TRUE.
   *
   * @return bool
   *   TRUE, if issues where found, FALSE otherwise.
   */
  public function hasIssues($run = TRUE);

  /**
   * Return all issues produced by this check.
   *
   * @return \pareviewsh\CheckIssue[]
   *   Array of issues produced by this check.
   */
  public function getIssues();

}
