<?php

namespace pareviewsh;

use \JsonSerializable;
use \pareviewsh\Exceptions\RuntimeException;

/**
 * Class for issues found while running a check.
 */
class CheckIssue extends Issue implements IssueInterface, FilesIssueInterface, IssueStatisticsInterface, JsonSerializable {

  /**
   * File issues.
   *
   * @var array<string, \pareviewsh\FileIssues>
   */
  protected $files;

  /**
   * Amount of notices.
   *
   * @var int
   */
  protected $notices;

  /**
   * Amount of warnings.
   *
   * @var int
   */
  protected $warnings;

  /**
   * Amount of errors.
   *
   * @var int
   */
  protected $errors;

  /**
   * {@inheritdoc}
   */
  public function addFileIssue(FileIssueInterface $file_issue) {
    $file_name = $file_issue->getFileName();

    // Whether we have a valid file name.
    if (empty($file_name)) {
      throw new RuntimeException('Could not determine the filename for the file issue.');
    }

    // Whether we don't have file issues for this file yet.
    if (!isset($this->files[$file_name])) {
      // Add a new file issues wrapper.
      $this->files[$file_name] = new FileIssues($file_name);
    }

    // Add the file issue.
    $this->files[$file_name]->addIssue($file_issue);

    // Adjust the parent issue severity, if the severity of the
    // file issue is higher than the severity of this issue.
    $severity = $file_issue->getSeverity();
    if (
      $this->severity === IssueInterface::SEVERITY_NONE
      || $severity === IssueInterface::SEVERITY_ERROR
    ) {
      $this->severity = $severity;
    }
    else if (
      $severity === IssueInterface::SEVERITY_WARNING
      && $this->severity !== IssueInterface::SEVERITY_ERROR
    ) {
      $this->severity = $severity;
    }
    else if (
      $severity === IssueInterface::SEVERITY_NOTICE
      && !in_array($this->severity, [
        IssueInterface::SEVERITY_ERROR,
        IssueInterface::SEVERITY_WARNING
      ], TRUE)
    ) {
        $this->severity = $severity;
    }

    switch ($severity) {
      case IssueInterface::SEVERITY_NOTICE:
        $this->notices++;
        break;
      case IssueInterface::SEVERITY_WARNING:
        $this->warnings++;
        break;
      case IssueInterface::SEVERITY_ERROR:
        $this->errors++;
        break;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getFileIssues() {
    return $this->files;
  }

  /**
   * {@inheritdoc}
   */
  public function hasFileIssues() {
    return count($this->files) > 0;
  }

  /**
   * {@inheritdoc}
   */
  public function getNoticesCount() {
    return $this->notices;
  }

  /**
   * {@inheritdoc}
   */
  public function getWarningsCount() {
    return $this->warnings;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrorsCount() {
    return $this->errors;
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return array_merge(parent::jsonSerialize(), [
      'files' => $this->files,
      'notices' => $this->getNoticesCount(),
      'warnings' => $this->getWarningsCount(),
      'errors' => $this->getErrorsCount(),
    ]);
  }

  /**
   * Reset all properties to their initial values.
   */
  protected function resetDefaults() {
    parent::resetDefaults();

    $this->files = [];
    $this->notices = 0;
    $this->warnings = 0;
    $this->errors = 0;
  }

}
