<?php

namespace pareviewsh;

/**
 * PAReview.sh checks.
 */
class Checks implements IssueStatisticsInterface, \Iterator, \JsonSerializable {

  /**
   * Available checks information.
   *
   * @var array<string, array>
   */
  protected static $availableChecks;

  /**
   * Enabled checks.
   *
   * @var array<string, \pareviewsh\CheckInterface>
   */
  protected $values;

  /**
   * Whether the enabled checks have been executed.
   *
   * @var bool
   */
  protected $executed;

  /**
   * Iterator index for enabled checks.
   *
   * @var int
   */
  protected $position;

  protected $notices;

  protected $warnings;

  protected $errors;

  /**
   * Return an array with available checks information.
   *
   * @return array
   *   Associative array keyed by check ID and a check information
   *   array as value that has the following keys:
   *   - id:                ID of the check.
   *   - class:             Class name.
   *   - name:              Human-readable name of the check.
   *   - short_description: Short description of the check.
   *   - long_description:  Long description of the check.
   *   - priority:          Priority of the check.
   */
  public static function getAvailableChecksInfo() {
    if (!isset(self::$availableChecks)) {
      $checks = [];

      $check_classes = array_values(array_filter(
        get_declared_classes(), 
        function ($class_name) {
          $testClass = new \ReflectionClass($class_name);
          return !$testClass->isAbstract() && $testClass->implementsInterface('pareviewsh\\CheckInterface');
        }
      ));

      foreach ($check_classes as $class_name) {
        $checks[$class_name::getId()] = [
          'id' => $class_name::getId(),
          'class' => $class_name,
          'name' => $class_name::getName(),
          'description' => $class_name::getDescription(),
          'priority' => $class_name::getPriority(),
          'requires_git' => $class_name::requiresGit(),
        ];
      }

      ksort($checks);

      self::$availableChecks = $checks;
    }

    return self::$availableChecks;
  }

  /**
   * Reset properties to their default values.
   */
  protected function resetDefaults() {
    $this->executed = FALSE;
    $this->position = 0;
    $this->notices = 0;
    $this->warnings = 0;
    $this->errors = 0;
  }

  /**
   * Enable a check.
   *
   * @param string $check_id
   *   Identifier of the check to enable.
   * @param bool $resort
   *   [Optional] Whether to resort enabled checks by priority,
   *   after the new check was enabled.
   *
   * @return bool
   *   TRUE, if the check was added successfully, FALSE if no
   *   changes to the enabled checks were done (e.g. when the
   *   check does not exist or was enabled already).
   */
  protected function enableCheck($check_id, $resort = TRUE) {
    // Whether the check does not exist or is already enabled.
    if (isset($this->values[$check_id]) || empty(self::$availableChecks[$check_id]['class'])) {
      return FALSE;
    }

    $class_name = self::$availableChecks[$check_id]['class'];
    try {
      $this->values[$check_id] = new $class_name();
    }
    catch (\Exception $ex) {
      $error = "ERROR: Cannot enable check \"{$check_id}\".";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $ex->getMessage();
      throw new \pareviewsh\Exceptions\RuntimeException($error, 3);
    }

    // Whether to resort the checks.
    if ($resort) {
      uasort($this->values, function($a, $b) {
        return $a::getPriority() - $b::getPriority();
      });
    }

    return TRUE;
  }

  /**
   * Constructs a checks instance.
   *
   * @param string|array $filter
   *   [Optional] Filter(s) to initialize the enabled checks.
   *   Defaults to '*' for enabling all available checks.
   *
   * @see self::getAvailableChecksInfo
   */
  public function __construct($filter = '*') {
    $this->values = [];
    $this->enableChecks($filter);
    $this->resetDefaults();
  }

  /**
   * Enable (disable) available checks by the given filter.
   *
   * The given filter(s) will be applied to the IDs of the available
   * checks.
   *
   * @param string|array $filters
   *   [Optional] Arry of filters or a filter string. A filter string may
   *   contain multiple filters separated by a comma (','). Filters can be
   *   inverted by using a '!' as prefix and use the wildcard placeholder
   *   '*' as suffix to apply to multiple filters with the given prefix.
   *   Defaults to '*' to include all available checks.
   *
   * @see self::getAvailableChecksInfo
   */
  public function enableChecks($filters = '*') {
    $available_checks = array_keys(self::getAvailableChecksInfo());

    // Whether no checks are available or the filters are empty.
    if (empty($available_checks) || empty($filters)) {
      // Disable all checks.
      $this->values = [];
      return;
    }

    // Whether the filters is a string.
    if (is_string($filters)) {
      // Split string by ',' to extract concatenated filters.
      $filters = array_map('trim', explode(',', $filters));
    }

    $matches = [];

    // Apply all filters.
    foreach($filters as $filter) {
      // Whether to invert the filter.
      $invert = FALSE;
      if (strpos($filter, '!') === 0) {
        $invert = TRUE;
        $filter = substr($filter, 1);
      }

      // Trim filter value.
      $filter = trim($filter);

      // Whether the filter is empty.
      if (empty($filter)) {
        $matches = $invert ? $available_checks : [];
      }
      // Whether the filter is our `all` placeholder.
      elseif ($filter == '*') {
        $matches = $invert ? [] : $available_checks;
      }
      // Whether the filter ends with our `all` placeholder.
      elseif (strpos($filter, '*') === strlen($filter) - 1) {
        $filter = preg_quote(substr($filter, 0, strlen($filter) - 1));
        $keys = preg_grep("/^{$filter}(.*)/", $available_checks);
        $matches = $invert ? array_diff($matches, $keys) : $matches + array_diff($keys, $matches);
      }
      // Exact match.
      elseif (in_array($filter, $available_checks)) {
        if ($invert) {
          // Array_search occasionally returned false keys, so we flip
          // keys and values, unset the key and take array_keys back.
          $keyed = array_flip($matches);
          unset($keyed[$filter]);
          $matches = array_keys($keyed);
        }
        elseif (!$invert && !in_array($filter, $matches)) {
          $matches[] = $filter;
        }
      }
    }

    // Remove all checks, that did not match.
    foreach (array_keys($this->values) as $key) {
      if (!in_array($key, $matches)) {
        unset($this->values[$key]);
      }
    }

    // Add new checks.
    $resort = FALSE;
    foreach ($matches as $key) {
      $resort = $this->enableCheck($key, FALSE) || $resort;
    }

    // Whether to resort the checks.
    if ($resort) {
      uasort($this->values, function($a, $b) {
        return $a::getPriority() - $b::getPriority();
      });
    }
  }

  /**
   * Disable checks by check status.
   *
   * This is a helper method that can be used to disable checks
   * that did not run after a previous check returned a run-time
   * error status.
   *
   * @param type $status
   */
  public function disableChecksByStatus($status) {
    $this->values = array_filter($this->values, function (CheckInterface $check) use($status) {
      return ($check->getStatus() !== $status);
    });
  }

  /**
   * Return the check information for the enabled checks.
   *
   * @return array
   *   Associative array keyed by check ID and a check information
   *   array as value that has the following keys:
   *   - id:                ID of the check.
   *   - class:             Class name.
   *   - name:              Human-readable name of the check.
   *   - short_description: Short description of the check.
   *   - long_description:  Long description of the check.
   *   - priority:          Priority of the check.
   */
  public function getEnabledChecksInfo() {
    return array_intersect_key(self::getAvailableChecksInfo(), $this->values);
  }

  /**
   * Prepare an all-checks run.
   *
   * Resets properties to their initial values and removes all Git
   * checks, if a local directory is to be investigated.
   */
  public function prepareRun() {
    $this->resetDefaults();
    $config = Pareview::getInstance()->getConfig();
    // Whether we check a local directory.
    if (!empty($config->directory)) {
      // Remove Git checks if we check a local folder.
      $this->values = array_filter($this->values, function (CheckInterface $check) {
        return (empty($check->requiresGit()));
      });
    }
  }

  /**
   * Executes all enabled checks.
   *
   * @return int
   *   Error code according to the issues identified:
   *   - 0: No issues or only notices found.
   *   - 1: Warnings found.
   *   - 2: Errors found.
   */
  public function run() {
    $this->prepareRun();

    foreach ($this->values as $check) {
      $status = $check->run();
      $this->notices += $check->getNoticesCount();
      $this->warnings += $check->getWarningsCount();
      $this->errors += $check->getErrorsCount();
      // Whether the check ended in an unknown status.
      if ($status === CheckInterface::CHECK_STATUS_UNKNOWN) {
        // We take an unknown status as indicator that
        // following checks should not run. So disable all
        // checks that did not run, and prevent them from
        // appearing in the resultset.
        $this->disableChecksByStatus(CheckInterface::CHECK_STATUS_NO_RUN);
        // Bail out.
        break;
      }
    }

    $this->executed = TRUE;

    // Whether we found errors.
    if ($this->errors > 0) {
      return 2;
    }
    // Whether we found warnings.
    else if ($this->warnings > 0) {
      return 1;
    }

    // On notices or no issues return 0.
    return 0;
  }

  /**
   * {@inheritdoc}
   */
  public function hasIssues() {
    return $this->notices + $this->warnings + $this->errors > 0;
  }

  /**
   * {@inheritdoc}
   */
  public function getNoticesCount() {
    return $this->notices;
  }

  /**
   * {@inheritdoc}
   */
  public function getWarningsCount() {
    return $this->warnings;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrorsCount() {
    return $this->errors;
  }

  /**
   * {@inheritdoc}
   */
  public function current() {
    return $this->values[$this->key()];
  }

  /**
   * {@inheritdoc}
   */
  public function key() {
    return array_keys($this->values)[$this->position];
  }

  /**
   * {@inheritdoc}
   */
  public function next() {
    ++$this->position;
  }

  /**
   * {@inheritdoc}
   */
  public function rewind() {
    $this->position = 0;
  }

  /**
   * {@inheritdoc}
   */
  public function valid() {
    return isset(array_keys($this->values)[$this->position]);
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return $this->values;
  }

}
