<?php

namespace pareviewsh\Checks;

use \JsonSerializable;
use \pareviewsh\CheckBase;
use \pareviewsh\CheckInterface;
use \pareviewsh\CheckIssue;
use \pareviewsh\Config;
use \pareviewsh\Exceptions\RuntimeException;
use \pareviewsh\IssueStatisticsInterface;
use \pareviewsh\Parser\CheckstyleXmlParser;
use \pareviewsh\Tools\CodeSniffer;
use \pareviewsh\Utils\MiscUtil;

/**
 * PAReview.sh Drupal coding standards check.
 *
 * Checks for compliance with Drupal coding standards. This check uses
 * PHP CodeSniffer and the Coder "Drupal" sniffs.
 */
class DrupalCodingStandardsCheck extends CheckBase implements CheckInterface, IssueStatisticsInterface, JsonSerializable {

  /**
   * {@inheritdoc}
   */
  public static function getId() {
    return 'drupal:coding_standards';
  }

  /**
   * {@inheritdoc}
   */
  public static function getPriority() {
    return 10;
  }

  /**
   * {@inheritdoc}
   */
  public static function getName() {
    return 'Drupal Coding Standards';
  }

  /**
   * {@inheritdoc}
   */
  public static function getDescription() {
    return "Checks for Drupal coding standards issues using Coder's \"Drupal\" inspection rules.";
  }

  /**
   * {@inheritdoc}
   */
  public function run() {
    $vendor_dir = Config::getVendorDir();
    $standards = Config::getConfigFileData('PHPCS_DRUPAL', realpath($vendor_dir . '/drupal/coder/coder_sniffer/Drupal'));

    // Whether the standards exist.
    if (!file_exists($standards)) {
      $error = sprintf("[%s] Coding standard \Drupal\" not found.", self::getId());
      throw new RuntimeException($error, 3);
    }
    // Whether PHP CS can be used.
    if (!CodeSniffer::canRun()) {
      $error = sprintf("[%s] PHP CodeSniffer (phpcs) not found.", self::getId());
      throw new RuntimeException($error, 3);
    }

    // Initialize PHPCS.
    $phpcs = new CodeSniffer($standards);
    $phpcs->setExtensions([]);
    $phpcs->run();

    // Get and parse PHPCS output.
    $output = $phpcs->getOutput();
    if (strpos($output, '<?xml') === 0) {
      $issues = CheckstyleXmlParser::parse($output, self::getId());
      if (!empty($issues)) {
        // Set general information.
        $files_issue = new CheckIssue([
          'message' => "The Coder \"Drupal\" inspection rules found some issues with your code.",
          'help_links' => [
            'Drupal Coding Standards' => 'https://www.drupal.org/node/318',
            'Coder' => 'https://www.drupal.org/project/coder',
          ],
         'test_id' => self::getId(),
        ]);
        // Add file specific issues.
        foreach ($issues as $issue) {
          $files_issue->addFileIssue($issue);
        }
        $files_issue->setMessage(sprintf("The Coder \"Drupal\" inspection rules found %s in your code.", MiscUtil::getSeverityCountList($files_issue->getNoticesCount(), $files_issue->getWarningsCount(), $files_issue->getErrorsCount())));
        // Add issue to check.
        $this->addIssue($files_issue);
      }
    }

    // Return status value based on the issues found.
    return parent::run();
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return array_merge(parent::jsonSerialize(), [
      'id' => self::getId(),
      'priority' => self::getPriority(),
      'name' => self::getName(),
      'description' => self::getDescription(),
    ]);
  }
}
