<?php

namespace pareviewsh\Checks;

use \JsonSerializable;
use \pareviewsh\CheckBase;
use \pareviewsh\CheckInterface;
use \pareviewsh\CheckIssue;
use \pareviewsh\Config;
use \pareviewsh\Exceptions\RuntimeException;
use \pareviewsh\IssueStatisticsInterface;
use \pareviewsh\Parser\CheckstyleXmlParser;
use \pareviewsh\Tools\EsLint;
use \pareviewsh\Utils\MiscUtil;

/**
 * PAReview.sh Drupal JavaScript Coding Standards check.
 *
 * Checks for compliance with Drupal JavaScript Coding Standards.
 * This check uses ESLint and the Drupal code ESLint rules.
 */
class DrupalJavaScriptCheck extends CheckBase implements CheckInterface, IssueStatisticsInterface, JsonSerializable {

  /**
   * {@inheritdoc}
   */
  public static function getId() {
    return 'drupal:javascript';
  }

  /**
   * {@inheritdoc}
   */
  public static function getPriority() {
    return 40;
  }

  /**
   * {@inheritdoc}
   */
  public static function getName() {
    return 'Drupal JavaScript Coding Standards';
  }

  /**
   * {@inheritdoc}
   */
  public static function getDescription() {
    return "Checks for Drupal JavaScript coding standards issues using ESLint inspection rules.";
  }

  /**
   * {@inheritdoc}
   */
  public function run() {
    $eslint = NULL;
    $config = realpath(Config::getPareviewRoot() . '/eslint.json');

    // Whether the standards exist.
    if (!file_exists($config)) {
      $error = sprintf("[%s] ESlint configuration not found.", self::getId());
      throw new RuntimeException($error, 3);
    }
    // Whether PHP CS can be used.
    if (!EsLint::canRun()) {
      $error = sprintf("[%s] ESlint not found.", self::getId());
      throw new RuntimeException($error, 3);
    }

    $eslint = new EsLint($config);
    $eslint->run();

    $output = $eslint->getOutput();
    if (strpos($output, '<?xml') === 0) {
      $issues = CheckstyleXmlParser::parse($output, self::getId());
      if (!empty($issues)) {
        // Set general information.
        $files_issue = new CheckIssue([
          'message' => 'ESLint found some issues with your code.',
          'help_links' => [
            'JavaScript Coding Standards' => 'https://www.drupal.org/node/172169',
            'ESLint' => 'https://eslint.org',
          ],
         'test_id' => self::getId(),
        ]);
        // Add file specific issues.
        foreach ($issues as $issue) {
          $files_issue->addFileIssue($issue);
        }
        $files_issue->setMessage(sprintf("ESLint found %s in your code.", MiscUtil::getSeverityCountList($files_issue->getNoticesCount(), $files_issue->getWarningsCount(), $files_issue->getErrorsCount())));
        // Add issue to check.
        $this->addIssue($files_issue);
      }
    }

    // Return status value based on the issues found.
    return parent::run();
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return array_merge(parent::jsonSerialize(), [
      'id' => self::getId(),
      'priority' => self::getPriority(),
      'name' => self::getName(),
      'description' => self::getDescription(),
    ]);
  }

}
