<?php

namespace pareviewsh\Checks;

use \JsonSerializable;
use \pareviewsh\CheckBase;
use \pareviewsh\CheckInterface;
use \pareviewsh\CheckIssue;
use \pareviewsh\Config;
use \pareviewsh\Exceptions\RuntimeException;
use \pareviewsh\IssueStatisticsInterface;
use \pareviewsh\Parser\CheckstyleXmlParser;
use \pareviewsh\Tools\CodeSniffer;
use \pareviewsh\Utils\MiscUtil;

/**
 * PAReview.sh Drupal Secure check.
 *
 * Checks for PHP CodeSniffer issues with the Drupal secure sniffs.
 */
class DrupalSecureCheck extends CheckBase implements CheckInterface, IssueStatisticsInterface, JsonSerializable {

  /**
   * {@inheritdoc}
   */
  public static function getId() {
    return 'drupal:secure';
  }

  /**
   * {@inheritdoc}
   */
  public static function getPriority() {
    return 30;
  }

  /**
   * {@inheritdoc}
   */
  public static function getName() {
    return 'Drupal Secure';
  }

  /**
   * {@inheritdoc}
   */
  public static function getDescription() {
    return "Checks for known security vulnerabilities using PHP Code Sniffer with \"DrupalSecure\" inspection rules.";
  }

  /**
   * {@inheritdoc}
   */
  public function run() {
    $vendor_dir = Config::getVendorDir();
    $standards = Config::getConfigFileData('PHPCS_DRUPALSECURE', realpath($vendor_dir . '/drupal/drupalsecure/DrupalSecure'));

    // Whether the standards exist.
    if (!file_exists($standards)) {
      $error = sprintf("[%s] Drupal secure sniffs not found.", self::getId());
      throw new RuntimeException($error, 3);
    }
    // Whether PHP CS can be used.
    if (!CodeSniffer::canRun()) {
      $error = sprintf("[%s] PHP CodeSniffer (phpcs) not found.", self::getId());
      throw new RuntimeException($error, 3);
    }

    // Initialize PHPCS.
    $phpcs = new CodeSniffer($standards);
    $phpcs->setExtensions([
      'php',
      'module',
      'inc',
      'install',
      'test',
      'profile',
      'theme',
      'yml',
    ]);
    $phpcs->run();

    $output = $phpcs->getOutput();
    if (strpos($output, '<?xml') === 0) {
      $issues = CheckstyleXmlParser::parse($output, self::getId());
      if (!empty($issues)) {
        // Set general information.
        $files_issue = new CheckIssue([
          'message' => "DrupalSecure found some issues with your code.",
          'help_links' => [
            'how to write secure code' => 'https://www.drupal.org/writing-secure-code',
            'DrupalSecure' => 'https://www.drupal.org/sandbox/coltrane/1921926',
          ],
         'test_id' => self::getId(),
        ]);
        // Add file specific issues.
        foreach ($issues as $issue) {
          $files_issue->addFileIssue($issue);
        }
        $files_issue->setMessage(sprintf("DrupalSecure found %s in your code.", MiscUtil::getSeverityCountList($files_issue->getNoticesCount(), $files_issue->getWarningsCount(), $files_issue->getErrorsCount())));
        // Add issue to check.
        $this->addIssue($files_issue);
      }
    }

    // Return status value based on the issues found.
    return parent::run();
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return array_merge(parent::jsonSerialize(), [
      'id' => self::getId(),
      'priority' => self::getPriority(),
      'name' => self::getName(),
      'description' => self::getDescription(),
    ]);
  }

}
