<?php

namespace pareviewsh;

use \pareviewsh\Exceptions\CliException;
use \pareviewsh\Exceptions\RuntimeException;
use \pareviewsh\Utils\Output;
use \pareviewsh\Utils\MiscUtil;

/**
 * Configuration for running PAReview.sh.
 *
 * Parses the command line to determine user supplied values.
 *
 * @property bool $isCli
 *   Whether PAReview.sh is running in command-line mode.
 * @property string $checks
 *   The checks filter string used to initialize the PAReview.sh
 *   checks.
 * @property string $directory
 *   The given location of a local project to investigate.
 * @property string $gitRepository
 *   The given Git repository URL of the remote project to
 *   investigate.
 * @property string $gitBranch
 *   The given Git branch of the remote project to investigate.
 * @property string $workspace
 *   The workspace directory of PAReview.sh to use for the
 *   investigation. If a local project directory is given,
 *   both properties will be the same. If a remote Git
 *   repository URL is given, it will be checked out to this
 *   workspace folder.
 * @property int $verbosity
 *   How verbose the output should be:
 *   - 0: No unnecessary output.
 *   - 1: Basic output for checks conducted.
 *   - 2: Shows checks conducted and files investigated.
 *   - 3: Detailed (debug) output.
 * @property bool $showColors
 *   Whether to enable colors in the generated report.
 * @property bool $showProofs
 *   Whether to include source code snippets showing the
 *   detected violations in the generated report.
 * @property bool $showProgress
 *   Whether to show on-screen progress when generating file
 *   reports.
 * @property bool quiet
 *   Whether quiet mode is enabled and screen output
 *   suppressed.
 * @property string $format
 *   The formatter to be used for generating the report.
 * @property string $reportFile
 *   Target file to write the report to.
 * @property string[] $unknown
 *   User specified command-line arguments, that are unknown to
 *   PAREView.sh.
 */
class Config {

  /**
   * The current PAReview.sh version.
   *
   * @var string
   */
  const VERSION = '2.x';

  /**
   * Package stability; either stable, beta, alpha or dev.
   *
   * @var string
   */
  const STABILITY = 'dev';

  /**
   * Vendor directory with test tools and coding standard sniffs.
   *
   * @var string
   */
  protected static $vendorDir;

  /**
   * Composer bin directory with test tools.
   *
   * @var string
   */
  protected static $binDir;

  /**
   * Config file content
   *
   * @var array<string, string>
   */
  protected static $configFile;

  /**
   * Whether to kill PAReview.sh on unknown command line arguments found.
   *
   * If FALSE, arguments that are not command line options will be ignored
   * and execution will continue. These values will be stored in
   * `$this->unknown`.
   *
   * @var bool
   */
  protected $dieOnUnknownArg;

  /**
   * Output handler
   *
   * @var \pareviewsh\Utils\Output
   */
  protected $output;

  /**
   * An array of settings that PAReview.sh uses.
   *
   * This array is not meant to be accessed directly. Instead, use the
   * settings as if they are class member vars so the __get() and __set()
   * magic methods can be used to validate the values.
   *
   * The list of settings are:
   * - string   checks        The checks being used for checking.
   * - string   directory     The directory to check.
   * - string   gitRepository The Git repository URL to check.
   * - string   gitBranch     The Git branch to check.
   * - string   workspace     The directory to check.
   * - int      verbosity     How verbose the output should be.
   *                          0: no unnecessary output
   *                          1: basic output for files being checked
   *                          2: ruleset and file parsing output
   *                          3: sniff execution output
   * - bool     showColors    Display colors in output.
   * - bool     showProofs    Extract code snippets showing the violations.
   * - bool     showProgress  Show basic progress information while running.
   * - bool     quiet         Quiet mode; disables progress and verbose output.
   * - string   format        The format of the results to use.
   * - string   reportFile    A file where the report output should be written.
   * - string[] unknown       Any arguments gathered on the command line that
   *                          are unknown to PAReview.sh.
   *
   * @var array<string, mixed>
   */
  private $settings = [
    'isCli' => null,
    'checks' => null,
    'directory' => null,
    'gitRepository' => null,
    'gitBranch' => null,
    'workspace' => null,
    'verbosity' => null,
    'showColors' => null,
    'showProofs' => null,
    'showProgress' => null,
    'quiet' => null,
    'format' => null,
    'reportFile' => null,
    'unknown' => null,
  ];

  /**
   * The current command line arguments we are processing.
   *
   * @var string[]
   */
  private $cliArgs = [];

  /**
   * Command line values that the user has supplied directly.
   *
   * @var array<string, TRUE>
   */
  private $overriddenDefaults = [];

  /**
   * Return a list of known source code formatters by file extension.
   *
   * @return array
   *   Associative array keyed by file extensions having according
   *   source code formatter identifiers as values.
   */
  public static function getFormattersByExtension() {
    return [
      'module' => 'php',
      'theme' => 'php',
      'php' => 'php',
      'profile' => 'php',
      'inc' => 'php',
      'install' => 'php',
      'js'  => 'javascript',
      'test' => 'php',
      'css' => 'css',
      'yml' => 'yaml',
      'twig' => 'twig',
      'tpl' => 'php',
      'md' => 'markdown',
    ];
  }

  /**
   * Return available report formatter descriptions.
   *
   * @return array<string, string>
   *   Associative array of available report formats with format
   *   identifiers as keys and short descriptions as values.
   */
  public static function getFormats() {
    return [
      'screen' => 'Textual report to be printed in the command line interface.',
      'classic' => 'Condensed report with minimal HTML formatting. Suitable for a comment in the drupal.org Project Applications issue queue.',
      //'html' => 'Full HTML report.',
      //'markdown' => 'GitHub flavored markdown report.',
      'json' => 'JSON formatted report.',
      'json-pretty' => 'Prettified JSON report.',
    ];
  }

  /**
   * Return PAReview.sh root path.
   *
   * @return string
   */
  public static function getPareviewRoot() {
    return realpath(__DIR__ . '/..');
  }

  /**
   * Return a property from the pareview.config file.
   *
   * The config file will be (re-)created on every Composer update and
   * install and holds the target paths of some of the executables and
   * standards used by PAReview.sh and installed by using Composer.
   *
   * @param string $property
   *   Identifier for the Property.
   * @param string|null $default
   *   Default value to be used if the value could not be found.
   *
   * @return string|null
   *   The config file value or the provided default value, if it was
   *   not found in the config file.
   */
  public static function getConfigFileData($property, $default = NULL) {
    if (!isset(self::$configFile) && file_exists(self::getPareviewRoot() . '/pareview.config')) {
      self::$configFile = parse_ini_file(self::getPareviewRoot() . '/pareview.config');
    }
    return !empty(self::$configFile[$property]) ? self::$configFile[$property] : $default;
  }

  /**
   * Return Composer's vendor directory.
   *
   * @return string
   *   Composer vendor directory.
   */
  public static function getVendorDir() {
    if (!isset(self::$vendorDir)) {
      $vendor_path = '';
      $pareview_root = self::getPareviewRoot();

      // Whether PAReview is the root package.
      if (file_exists(realpath($pareview_root . '/vendor/autoload.php'))) {
        $vendor_path = realpath($pareview_root . '/vendor');
      }
      else {
        // Traverse the current directory three levels up
        // and try finding composer's autoload.php.
        $path = realpath($pareview_root . '/..');
        for ($i = 0; $i <= 3; $i++) {
          if (file_exists(realpath($path . '/autoload.php'))) {
            $vendor_path = $path;
            break;
          }
          $new_path = realpath($path . '/..');
          if ($new_path == $path) {
            break;
          }
          else {
            $path = $new_path;
          }
        }
      }

      self::$vendorDir = self::getConfigFileData('COMPOSER_VENDOR', $vendor_path);
    }

    return self::$vendorDir;
  }

  /**
   * Return Composer's bin directory.
   *
   * @return string
   *   Composer bin directory.
   */
  public static function getBinDir() {
    if (!isset(self::$binDir)) {
      $vendor_path = self::getVendorDir();
      $bin_dir = '';

      if (file_exists($vendor_path . '/bin') && is_dir($vendor_path . '/bin')) {
        $bin_dir = realpath($vendor_path . '/bin');
      }

      self::$binDir = self::getConfigFileData('COMPOSER_BIN', $bin_dir);
    }

    return self::$binDir;
  }

  /**
   * Creates a Config object and populates it with command
   * line values.
   *
   * @param \pareviewsh\Utils\Output
   *   Output handler.
   * @param bool $is_cli
   *   Whether PAReview.sh is running in CLI mode.
   * @param array $cliArgs
   *   An array of values gathered from CLI arguments.
   * @param bool $dieOnUnknownArg
   *   Whether to kill the process when an unknown command
   *   line argument is found.
   */
  public function __construct(Output $output, $is_cli = TRUE, array $cliArgs = [], $dieOnUnknownArg = TRUE) {
    $this->output = $output;
    $this->isCli = $is_cli;

    // Restore settings values.
    $this->restoreDefaults();

    if (empty($cliArgs) === true) {
      $cliArgs = $_SERVER['argv'];
      array_shift($cliArgs);
    }

    if ($is_cli && empty($cliArgs)) {
      $error = "ERROR: Missing arguments.";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
    $this->setCommandLineValues($cliArgs);
    if (
      $is_cli
      && !isset($this->overriddenDefaults['directory'])
      && !isset($this->overriddenDefaults['gitRepository'])
    ) {
      $error = "ERROR: You must either provide a local path to review or specify a Git repository using the \"--git\" option.";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
  }

  /**
   * Get the value of a setting.
   *
   * @param string $name
   *   The name of the property.
   *
   * @return mixed
   *   Value of the setting.
   *
   * @throws \pareviewsh\RuntimeException
   *   If the setting name is invalid.
   */
  public function __get($name) {
    if (array_key_exists($name, $this->settings) === false) {
      throw new RuntimeException("Unable to get value of setting \"$name\".");
    }
    return $this->settings[$name];
  }

  /**
   * Set the value of a setting.
   *
   * @param string $name
   *   The name of the setting.
   * @param mixed $value
   *   The value of the setting.
   *
   * @throws \pareviewsh\RuntimeException
   *   If the setting name is invalid.
   */
  public function __set($name, $value) {
    if (array_key_exists($name, $this->settings) === false) {
      throw new RuntimeException("Can't set value of \"$name\"; setting doesn't exist.");
    }
    $this->settings[$name] = $value;
  }

  /**
   * Whether a setting value was set.
   *
   * @param string $name
   *   The name of the setting.
   *
   * @return bool
   *   TRUE, if the setting is set, FALSE otherwise.
   */
  public function __isset($name) {
    return isset($this->settings[$name]);
  }

  /**
   * Unset a setting value.
   *
   * @param string $name
   *   The name of the setting.
   */
  public function __unset($name) {
    $this->settings[$name] = null;
  }

  /**
   * Return all PAReview.sh settings as array.
   *
   * @return array<string, mixed>
   *   Associative array of PAReview.sh settings with the setting names
   *   as keys and their values as values.
   */
  public function getSettings() {
    return $this->settings;
  }

  /**
   * Set the array of all settings.
   *
   * @param array<string, mixed> $settings
   *   An associative array of PAReview.sh settings with the setting
   *   names as keys and their values as values.
   */
  public function setSettings($settings) {
    return $this->settings = $settings;
  }

  /**
   * Set the command line values.
   *
   * @param array $args
   *   An array of command line arguments to set.
   */
  public function setCommandLineValues(array $args) {
    $this->cliArgs = $args;
    $numArgs = count($args);
    for ($i = 0; $i < $numArgs; $i++) {
      $arg = $this->cliArgs[$i];
      if ($arg === '') {
        continue;
      }
      if ($arg{0} === '-') {
        if (in_array($arg, ['-', '--'])) {
          // Empty argument, ignore it.
          continue;
        }
        if ($arg{1} === '-') {
          $this->processLongArgument(substr($arg, 2), $i);
        }
        else {
          $switches = str_split($arg);
          foreach ($switches as $switch) {
            if ($switch === '-') {
              continue;
            }
            $this->processShortArgument($switch, $i);
          }
        }
      }
      else {
        $this->processUnknownArgument($arg, $i);
      }
    }
  }

  /**
   * Restore default values for all possible command line arguments.
   *
   * @return array
   */
  public function restoreDefaults() {
    $this->checks = '*';
    $this->directory = NULL;
    $this->gitRepository = NULL;
    $this->gitBranch = NULL;
    $this->workspace = realpath(self::getPareviewRoot() . '/pareview_temp');
    $this->verbosity = 0;
    $this->showColors = FALSE;
    $this->showProofs = FALSE;
    $this->showProgress = FALSE;
    $this->quiet = FALSE;
    $this->format = 'screen';
    $this->reportFile = NULL;
    $this->unknown = [];
  }

  /**
   * Process short (single letter) command line argument.
   *
   * @param string $arg
   *   The command line argument.
   * @param int $pos
   *   The position of the argument on the command line.
   */
  public function processShortArgument($arg, $pos) {
    switch ($arg) {
      // Help/usage.
      case 'h':
      case '?':
        ob_start();
        $this->printUsage();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // List checks.
      case 'l':
        ob_start();
        $this->printAvailableChecks();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // List formats.
      case 'f':
        ob_start();
        $this->printAvailableFormats();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // Verbose level.
      case 'v':
        if ($this->quiet === TRUE) {
          // Ignore when quiet mode is enabled.
          break;
        }
        $this->verbosity++;
        $this->overriddenDefaults['verbosity'] = TRUE;
        break;

      // Show source code snippets.
      case 's':
        $this->showProofs = TRUE;
        $this->overriddenDefaults['showProofs'] = TRUE;
        break;

      // Show progress.
      case 'p' :
        if ($this->quiet === TRUE) {
          // Ignore when quiet mode is enabled.
          break;
        }
        $this->showProgress = TRUE;
        $this->overriddenDefaults['showProgress'] = TRUE;
        break;

      // Enable colors.
      case 'c' :
        $this->showColors = true;
        $this->overriddenDefaults['showColors'] = true;
        break;

      // Enable quiet mode.
      case 'q' :
        // Quiet mode disables a few other settings as well.
        $this->quiet = true;
        if ($this->showProgress) {
          $this->showProgress = FALSE;
          unset($this->overriddenDefaults['showProgress']);
        }
        if ($this->verbosity) {
          $this->verbosity = 0;
          unset($this->overriddenDefaults['verbosity']);
        }
        $this->overriddenDefaults['quiet'] = TRUE;
        break;

      // Any other (unknown) switches.
      default:
        if ($this->dieOnUnknownArg === FALSE) {
          $unknown = $this->unknown;
          $unknown[] = $arg;
          $this->unknown = $unknown;
        }
        else {
          $this->processUnknownArgument('-' . $arg, $pos);
        }
        break;

    }
  }

  /**
   * Processes long (multi-letter) command line argument.
   *
   * @param string $arg
   *   The command line argument.
   * @param int $pos
   *   The position of the argument on the command line.
   */
  public function processLongArgument($arg, $pos) {
    switch ($arg) {
      // Help/usage.
      case 'help':
        ob_start();
        $this->printUsage();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // Version information.
      case 'version':
        ob_start();
        $this->printVersion();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // List checks.
      case 'list-checks':
        ob_start();
        $this->printAvailableChecks();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // List formats.
      case 'formats':
        ob_start();
        $this->printAvailableFormats();
        $output = ob_get_contents();
        ob_end_clean();
        throw new CliException($output, 0);

      // Show source code snippets.
      case 'source':
        if (isset($this->overriddenDefaults['showProofs']) === TRUE) {
          break;
        }
        $this->showProofs = TRUE;
        $this->overriddenDefaults['showProofs'] = TRUE;
        break;

      // Show progress.
      case 'progress':
        if ($this->quiet === TRUE || isset($this->overriddenDefaults['showProgress']) === TRUE) {
          // Ignore when quiet mode is enabled.
          break;
        }
        $this->showProgress = TRUE;
        $this->overriddenDefaults['showProgress'] = TRUE;
        break;

      // Enable colors.
      case 'colors':
        if (isset($this->overriddenDefaults['showColors']) === TRUE) {
          break;
        }
        $this->showColors = TRUE;
        $this->overriddenDefaults['showColors'] = TRUE;
        break;

      // Options.
      default:
        // Git repository URL.
        if (substr($arg, 0, 4) === 'git=') {
          if (isset($this->overriddenDefaults['gitRepository']) === TRUE) {
            break;
          }
          if (isset($this->overriddenDefaults['directory']) === TRUE) {
            $error = "ERROR: Option \"--git\" cannot be used together with a path argument.";
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
          $this->overriddenDefaults['gitRepository'] = TRUE;
          $this->gitRepository = substr($arg, 4);
          if (
            empty($this->gitRepository)
            || strpos($this->gitRepository, 'http') !== 0
            || filter_var($this->gitRepository, FILTER_SANITIZE_URL) === FALSE
          ) {
            $error = 'ERROR: The specified Git repository "' . $this->gitRepository . '" is not a valid URL.';
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
        }
        // Branch.
        else if (substr($arg, 0, 7) === 'branch=') {
          if (isset($this->overriddenDefaults['gitBranch']) === TRUE) {
            break;
          }
          $this->overriddenDefaults['gitBranch'] = TRUE;
          $this->gitBranch = substr($arg, 7);
          if (empty($this->gitBranch)) {
            $error = 'ERROR: The specified Git branch is not valid.';
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
        }
        // Workspace.
        else if (substr($arg, 0, 10) === 'workspace=') {
          if (isset($this->overriddenDefaults['workspace']) === TRUE) {
            break;
          }
          if (isset($this->overriddenDefaults['directory']) === TRUE) {
            $error = "ERROR: Option \"--workspace\" cannot be used together with a path argument.";
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
          $this->overriddenDefaults['workspace'] = TRUE;
          $this->workspace = substr($arg, 10);
          // @todo: validate workspace.
//          if (empty($this->gitBranch)) {
//            $error = 'ERROR: The specified Git branch is not valid.';
//            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(true);
//            throw new CliException($error, 3);
//          }
        }
        // Report format.
        else if (substr($arg, 0, 7) === 'format=') {
          if (isset($this->overriddenDefaults['format']) === TRUE) {
            break;
          }
          $this->overriddenDefaults['format'] = TRUE;
          $this->format = substr($arg, 7);
          if (!in_array($this->format, array_keys(self::getFormats()))) {
            $error = "ERROR: The given report format \"{$this->format}\" is not supported.";
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
        }
        // Report file.
        else if (substr($arg, 0, 12) === 'report-file=') {
          if (isset($this->overriddenDefaults['reportFile']) === TRUE) {
            break;
          }
          $this->overriddenDefaults['reportFile'] = TRUE;
          $this->reportFile = substr($arg, 12);
          $dir = dirname($this->reportFile);
          if (!file_exists($dir) || !is_dir($dir) || !is_writable($dir)) {
            $error = "ERROR: The path \"{$dir}\" for the report file does not exist or is not writable.";
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
          else if (file_exists($this->reportFile)) {
            $error = "ERROR: The report file \"{$this->reportFile}\" already exists and would be overwritten.";
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
            throw new CliException($error, 3);
          }
        }
        // Checks.
        else if (substr($arg, 0, 7) === 'checks=') {
          if (isset($this->overriddenDefaults['checks']) === TRUE) {
            break;
          }
          $this->checks = substr($arg, 7);
          $checks = Pareview::getInstance()->getChecks();
          $checks->enableChecks($this->checks);
          if (count($checks->getEnabledChecksInfo()) < 1) {
            $error = "ERROR: The given value \"{$this->checks}\" for option \"--checks\" does not match any check.";
            $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . "Run \"pareview -f\" to list all supported formats.";
            throw new CliException($error, 3);
          }
          $this->overriddenDefaults['checks'] = TRUE;
        }
        else {
          if ($this->dieOnUnknownArg === FALSE) {
            $eqPos = strpos($arg, '=');
            try {
              if ($eqPos === FALSE) {
                $this->values[$arg] = $arg;
              }
              else {
                $value = substr($arg, ($eqPos + 1));
                $arg = substr($arg, 0, $eqPos);
                $this->values[$arg] = $value;
              }
            }
            catch (RuntimeException $e) {
              // Value is not valid, so just ignore it.
            }
          }
          else {
            $this->processUnknownArgument('--' . $arg, $pos);
          }
        }
        break;

    }
  }

  /**
   * Processes an unknown command line argument.
   *
   * Assumes all unknown arguments are files and folders to check.
   *
   * @param string $arg
   *   The command line argument.
   * @param int $pos
   *   The position of the argument on the command line.
   */
  public function processUnknownArgument($arg, $pos) {
    // We don't know about any additional switches; just files.
    if ($arg{0} === '-') {
      if ($this->dieOnUnknownArg === FALSE) {
        return;
      }
      $error = "ERROR: Option \"$arg\" not known.";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }

    if (isset($this->overriddenDefaults['directory']) === TRUE) {
      if ($this->dieOnUnknownArg === FALSE) {
        return;
      }
      $error = "ERROR: Argument \"$arg\" unknown.";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
    else if (isset($this->overriddenDefaults['gitRepository']) === TRUE) {
      if ($this->dieOnUnknownArg === FALSE) {
        return;
      }
      $error = "ERROR: Argument \"$arg\" invalid when using \"--git\" option.";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
    else if (isset($this->overriddenDefaults['workspace']) === TRUE) {
      if ($this->dieOnUnknownArg === FALSE) {
        return;
      }
      $error = "ERROR: Argument \"$arg\" invalid when using \"--workspace\" option.";
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
    else {
      $this->processFilePath($arg);
    }
  }

  /**
   * Process a file path and add it as base directory.
   *
   * @param string $path
   *   The path to add.
   */
  public function processFilePath($path) {
    $dir = realpath($path);
    if (file_exists($dir) === FALSE) {
      if ($this->dieOnUnknownArg === FALSE) {
        return;
      }
      $error = 'ERROR: The path "' . $path . '" does not exist.';
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
    else if(is_dir($dir) !== TRUE) {
      if ($this->dieOnUnknownArg === FALSE) {
        return;
      }
      $error = 'ERROR: The path "' . $path . '" is not a directory.';
      $error = $this->output->format($error)->bg('red')->bold()->white() . PHP_EOL . PHP_EOL . $this->printShortUsage(TRUE);
      throw new CliException($error, 3);
    }
    else {
      $this->overriddenDefaults['directory'] = TRUE;
      $this->directory = $this->workspace = $dir;
    }
  }

  /**
   * Print PAReview.sh version.
   */
  public function printVersion() {
    echo $this->output->format('PAReview.sh version ' . self::VERSION . '-' . self::STABILITY)->bold()->light_blue() . ' ';
    echo '(https://www.drupal.org/project/pareviewsh)' . PHP_EOL;
  }

  /**
   * Print available checks.
   */
  public function printAvailableChecks() {
    $this->printVersion();
    echo PHP_EOL;
    echo MiscUtil::wrapLn('The "--checks" option allows enabling/disabling certain checks for the review. The wildcard suffix "*" can be used to enable all checks of a certain kind, while using the prefix "!" inverts the check pattern and disables the given check(s). Multiple check patterns can be concatenated using a comma ",". They will be applied in their order of appearance.', 0, 2);
    echo PHP_EOL;
    echo $this->output->format(MiscUtil::wrapLn('Examples:', 0, 2))->bold()->light_blue();
    echo MiscUtil::wrapLn('Enable all checks (default value):', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview --checks=' . escapeshellarg('*'). ' <folder>', 4, 2))->light_green() . PHP_EOL . PHP_EOL;
    echo MiscUtil::wrapLn('Enable all Drupal checks:', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview --checks=' . escapeshellarg('drupal:*'). ' <folder>', 4, 2))->light_green() . PHP_EOL . PHP_EOL;
    echo MiscUtil::wrapLn('Enable all Drupal checks, except the best practices check:', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview --checks=' . escapeshellarg('drupal:*,!drupal:practice') . ' <folder>', 4, 2))->light_green() . PHP_EOL . PHP_EOL;
    echo PHP_EOL;

    echo $this->output->format(MiscUtil::wrapLn('Available checks:', 0, 2))->bold()->light_blue();
    // Get available checks information.
    $available_checks = Checks::getAvailableChecksInfo();
    // Find longest identifier.
    $lengths = array_map('strlen', array_keys($available_checks));
    $max_length = max($lengths);
    $prefix = str_repeat(' ', $max_length + 4);

    // Print checks.
    foreach ($available_checks as $check) {
      echo "  " . $this->output->format(str_pad($check['id'], $max_length + 2))->light_green();
      // Print wrapped details.
      echo $this->output->format(MiscUtil::wrapLn($check['name'], 0, 2, 0, $prefix))->bold();
      echo $prefix . MiscUtil::wrapLn($check['description'], 0, 2, 0, $prefix) . PHP_EOL;
    }

    echo PHP_EOL;
  }

  /**
   * Print available report formatters.
   */
  public function printAvailableFormats() {
    $this->printVersion();
    echo PHP_EOL;
    echo MiscUtil::wrapLn('The default report format used by PAReview.sh is "screen", which is optimized for printing the review results to the currently used command line interface. The "--format" option allows switching to another supported report format.', 0, 2);
    echo PHP_EOL;

    echo $this->output->format(MiscUtil::wrapLn('Available formats:', 0, 2))->bold()->light_blue();

    $formats = self::getFormats();

    $lengths = array_map('strlen', array_keys($formats));
    $max_length = max($lengths);
    $prefix = str_repeat(' ', $max_length + 4);

    // Print formats.
    foreach ($formats as $key => $format) {
      echo "  " . $this->output->format(str_pad($key, $max_length + 2))->light_green();
      echo MiscUtil::wrapLn($format, 0, 2, 0, $prefix);
    }

    echo PHP_EOL;
  }

  /**
   * Prints out the short usage information for this script.
   *
   * @param bool $return
   *   If TRUE, the usage string is returned
   *   instead of output to screen.
   *
   * @return string|void
   */

  public function printShortUsage($return = false) {
    $usage = 'Run "pareview --help" for usage information.';
    $usage .= PHP_EOL . PHP_EOL;
    if ($return === true) {
      return $usage;
    }
    echo $usage;
  }

  /**
   * Prints out the usage information for PHPCS.
   *
   * @return void
   */
  public function printUsage() {
    $this->printVersion();
    echo PHP_EOL;
    echo MiscUtil::wrapLn('PAReview.sh is a command line tool to automatically review Drupal.org project applications. It either takes a Git repository URL as argument and clones its code into a local workspace, or it takes a local path to a module/theme project. The source code given will be reviewed for issues regarding Coding Standards, application of best practices, security issues, and other common flaws.', 0, 2);
    echo PHP_EOL;
    echo $this->output->format(MiscUtil::wrapLn('Usage:', 0, 2))->bold()->light_blue();
    echo $this->output->format(MiscUtil::wrapCommand('pareview [[switches]] [--git=<git-repository>] [--branch=<git-branch>] [--checks=<checks>] [--format=<format>] [--report-file=<report-file>] [--workspace=<workspace>] [<folder>]', 2, 2))->light_green() . PHP_EOL;
    echo PHP_EOL;
    echo $this->output->format(MiscUtil::wrapLn('Examples:', 0, 2))->bold()->light_blue();
    echo MiscUtil::wrapLn('Check the 8.x-1.x branch of the rules module in the remote Git repository of drupal.org:', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview --git=https://git.drupal.org/project/rules --branch=8.x-1.x', 4, 2))->light_green() . PHP_EOL . PHP_EOL;
    $example_path = (DIRECTORY_SEPARATOR == '\\') ? 'C:\Projects\mymodule' : '/projects/mymodule';
    echo MiscUtil::wrapLn('Check the local custom module at the path "' . $example_path . '":', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview ' . escapeshellarg($example_path), 4, 2))->light_green() . PHP_EOL . PHP_EOL;
    echo MiscUtil::wrapLn('Check only coding standards and best practices on the local custom module at the path "' . $example_path . '":', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview --checks="drupal:coding_standards,drupal:practice" ' . escapeshellarg($example_path), 4, 2))->light_green() . PHP_EOL . PHP_EOL;
    echo MiscUtil::wrapLn('Print colored results including source code snippets showing the violations identified:', 2, 2);
    echo $this->output->format(MiscUtil::wrapCommand('pareview -cs ' . escapeshellarg($example_path), 4, 2))->light_green() . PHP_EOL . PHP_EOL;

    $prefix = str_repeat(' ', 22);
    echo PHP_EOL;
    echo $this->output->format(MiscUtil::wrapLn('Switches:', 0, 2))->bold()->light_blue();

    echo '  ' . $this->output->format('-l')->bold()->light_green() . ' / ' . $this->output->format('--list-checks')->bold()->light_green() . '  ';
    echo MiscUtil::wrapLn('List all available checks.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-s')->bold()->light_green() . ' / ' . $this->output->format('--source')->bold()->light_green() . '       ';
    echo MiscUtil::wrapLn('Include source code snippets in results.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-f')->bold()->light_green() . ' / ' . $this->output->format('--formats')->bold()->light_green() . '      ';
    echo MiscUtil::wrapLn('List all available report formats.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-c')->bold()->light_green() . ' / ' . $this->output->format('--colors')->bold()->light_green() . '       ';
    echo MiscUtil::wrapLn('Enable report colors in textual report (experimental).', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-p')->bold()->light_green() . ' / ' . $this->output->format('--progress')->bold()->light_green() . '     ';
    echo MiscUtil::wrapLn('Show progress.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-q')->bold()->light_green() . ' / ' . $this->output->format('--quiet')->bold()->light_green() . '        ';
    echo MiscUtil::wrapLn('Quiet mode; disables progress and verbose output.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-h')->bold()->light_green() . ' / ' . $this->output->format('--help')->bold()->light_green() . '         ';
    echo MiscUtil::wrapLn('Print this help message.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('--version')->bold()->light_green() . '           ';
    echo MiscUtil::wrapLn('Print version information.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-v')->bold()->light_green() . '                  ';
    echo MiscUtil::wrapLn('Print processed checks.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-vv')->bold()->light_green() . '                 ';
    echo MiscUtil::wrapLn('Print processed checks and issues.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('-vvv')->bold()->light_green() . '                ';
    echo MiscUtil::wrapLn('Print verbose processing information.', 0, 2, 0, $prefix);

    echo PHP_EOL;
    echo $this->output->format(MiscUtil::wrapLn('Options:', 0, 2))->bold()->light_blue();

    echo '  ' . $this->output->format('--git')->bold()->light_green() . '               ';
    echo MiscUtil::wrapLn('Review the given Git repository. <git-repository> must be a publicly accessible Git checkout URL.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('--branch')->bold()->light_green() . '            ';
    echo MiscUtil::wrapLn('An optional branch to review when using the "--git" option. If not given, the default branch of the repository will be used.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('--checks')->bold()->light_green() . '            ';
    echo MiscUtil::wrapLn('A comma separated list of checks to conduct. The "-l" switch lists all available checks.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('--format')->bold()->light_green() . '            ';
    echo MiscUtil::wrapLn('Specify the report format. Defaults to "classic". The "-f" switch lists all available formats.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('--report-file')->bold()->light_green() . '       ';
    echo MiscUtil::wrapLn('Write the report to the specified file instead of printing it to the screen.', 0, 2, 0, $prefix);

    echo '  ' . $this->output->format('--workspace')->bold()->light_green() . '         ';
    echo MiscUtil::wrapLn('Target folder to checkout the Git repository to when using the "--git" option.', 0, 2, 0, $prefix);

    echo PHP_EOL;
    echo $this->output->format(MiscUtil::wrapLn('Optional argument:', 0, 2))->bold()->light_blue();

    echo '  ' . $this->output->format('<directory>')->bold()->light_green() . '         ';
    echo MiscUtil::wrapLn('A local folder to review. This argument may be used instead of the "--git" option to review local files only.', 0, 2, 0, $prefix);

    echo PHP_EOL;
  }

}
