<?php

namespace pareviewsh;

use \JsonSerializable;
use \pareviewsh\Exceptions\RuntimeException;

/**
 * File related issue class.
 */
class FileIssue extends Issue implements FileIssueInterface, JsonSerializable {

  /**
   * File path and name of the related source code file.
   *
   * This path is relative to the workspace of PAReview.sh.
   *
   * @var string
   */
  protected $file;

  /**
   * Source code line where the issue was found.
   *
   * This value is optional and may be NULL, if the issue is not directly
   * related to a source code line.
   *
   * @var int|null
   */
  protected $line;

  /**
   * Source code column (character) where the issue was found.
   *
   * This value is optional and may be NULL, if the issue is not directly
   * related to a source code column.
   *
   * @var int|null
   */
  protected $column;

  /**
   * Source code snippet with a visual representation of the issue.
   *
   * @var array
   */
  protected $proof;

  /**
   * {@inheritdoc}
   */
  public function __construct($issue = NULL, $extract_proof = FALSE) {
    parent::__construct($issue);

    // Whether to extract the proof from source file.
    if ($extract_proof) {
      $this->extractProof();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function __set($name, $value) {
    $property = $this->getPropertyName($name);
    if (!property_exists($this, $property)) {
      throw new RuntimeException("Unable to set value of property \"{$name}\" ({$property}).");
    }
    if ($property === 'fileName') {
      $this->setFileName($value);
    }
    else {
      $this->{$property} = $value;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setFileName($file_name, $strip_base = TRUE) {
    // By default, we store file names relative to PAReview.sh
    // base dir. So remove it from file name, if the file name
    // starts with the base dir.
    $base = Pareview::getInstance()->getWorkspaceDir();
    if ($strip_base && strpos($file_name, $base) === 0) {
      $file_name = '.' . substr($file_name, strlen($base));
    }

    $this->file = $file_name;
  }

  /**
   * {@inheritdoc}
   */
  public function getFileName() {
    return $this->file;
  }

  /**
   * {@inheritdoc}
   */
  public function setLine($line = NULL) {
    $this->line = $line;
  }

  /**
   * {@inheritdoc}
   */
  public function getLine() {
    return $this->line;
  }

  /**
   * {@inheritdoc}
   */
  public function setColumn($column = NULL) {
    $this->column = $column;
  }

  /**
   * {@inheritdoc}
   */
  public function getColumn() {
    return $this->column;
  }

  /**
   * {@inheritdoc}
   */
  public function extractProof() {
    // Unset any previously extracted proof.
    $this->proof = [];

    // Whether we have no file name or line.
    if (empty($this->file) || is_null($this->line)) {
      return FALSE;
    }

    // Get real fine name from base dir and self::fileName property.
    $base = Pareview::getInstance()->getBaseDir();
    $real_name = realname($base . '/' . $this->file);

    // Whether the file does not exist.
    if (!file_exists($real_name)) {
      return FALSE;
    }

    // Extract source code snippet from file.
    $file = new SourceFile(realname($base . '/' . $this->file));
    $this->proof = $file->extractSource($this->line);
    $file = null;

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function setProof(array $proof) {
    $this->proof = $proof;
  }

  /**
   * {@inheritdoc}
   */
  public function getProof() {
    return $this->proof;
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return array_merge(parent::jsonSerialize(), [
      'file' => $this->file,
      'line' => $this->line,
      'column' => $this->column,
      'proof' => $this->proof,
    ]);
  }

}
