<?php

namespace pareviewsh;

/**
 * Interface for file related issues.
 *
 * Additionally to the properties of a general issue, file related
 * issues contain information about the file where the issue was
 * identified.
 */
interface FileIssueInterface extends IssueInterface {

  /**
   * Construct an instance of an issue.
   *
   * @param object|array|null $issue
   *   [Optional] Initial issue values. Defaults to NULL and may be an
   *   empty value or a standard object/associative array with some or
   *   all of the following properties/keys:
   *   - file:       Source code file name.
   *   - line:       Issue-related source code line index.
   *   - column:     Issue-related source code character position.
   *   - message:    Human-readable issue message.
   *   - help_links: Associative array of issue help links.
   *   - severity:   Severity of the issue.
   *   - test_id:    ID of the related test.
   *   - rule_id:    ID of the related rule.
   *   - proof:      Source code snippet prooving the issue.
   * @param bool $extract_proof
   *   [Optional] Whether to extract source code proof (TRUE) or
   *   not (FALSE).
   */
  public function __construct($issue = NULL, $extract_proof = FALSE);

  /**
   * Set file path and name of the related source code file.
   *
   * @param string $file_name
   *   Source code file path/name.
   * @param bool $strip_base
   *   [Optional] Whether to strip the PAReview.sh workspace
   *   prefix from the file path (TRUE) or not (FALSE).
   *   Defaults to TRUE.
   */
  public function setFileName($file_name, $strip_base = TRUE);

  /**
   * Return file path and name of the related source code file.
   *
   * @return string
   *   File path/name of the source code file. This path is relative
   *   to the workspace of PAReview.sh.
   */
  public function getFileName();

  /**
   * Set source code line index where the issue was found.
   *
   * @param int|null $line
   *   [Optional] Source code line index.
   *   Defaults to NULL.
   */
  public function setLine($line = NULL);

  /**
   * Return source code line index where the issue was found.
   *
   * @return int|null
   *   Source code line index or NULL, if the issue is not
   *   directly related to a source code line/position.
   */
  public function getLine();

  /**
   * Set source code column (character) position where the issue was found.
   *
   * @param int|null $column
   *   [Optional] Source code column (character) position.
   *   Defaults to NULL.
   */
  public function setColumn($column = NULL);

  /**
   * Return source code column (character) position where the issue was found.
   *
   * @return int|null
   *   Source code column (character) position or NULL, if the issue is
   *   not directly related to a source code line/position.
   */
  public function getColumn();

  /**
   * Extract a snippet from the source code file showing the issue.
   *
   * @return bool
   *   TRUE on success, FALSE otherwise.
   */
  public function extractProof();

  /**
   * Set source code snippet with a visual representation of the issue.
   *
   * @param array $proof
   *   Empty array to delete a proof, or an associative array with the
   *   following keys:
   *   - start:     Start line of the snippet.
   *   - format:    Source code format (for syntax highlighters).
   *   - lines:     Array of the extracted lines.
   *   - eof:       Boolean indicating whether the snippet ends at end of file.
   */
  public function setProof(array $proof);

  /**
   * Return source code snippet with a visual representation of the issue.
   *
   * @return array
   *   Empty array, if no source code snippet was set or an associative
   *   array with the following keys:
   *   - start:     Start line of the snippet.
   *   - format:    Source code format (for syntax highlighters).
   *   - lines:     Array of the extracted lines.
   *   - eof:       Boolean indicating whether the snippet ends at end of file.
   */
  public function getProof();

}
