<?php

namespace pareviewsh;

use \JsonSerializable;

/**
 * File issues class.
 *
 * This class wraps issues related to a certain file and provides
 * methods for handling file information.
 */
class FileIssues implements IssuesInterface, JsonSerializable {

  /**
   * Issues for this file.
   *
   * @var FileIssueInterface[]
   */
  protected $issues;

  /**
   * File name.
   *
   * @var type 
   */
  protected $fileName;

  /**
   * Formatter for this type of file.
   *
   * @var string
   */
  protected $formatter;

  /**
   * Amount of notices for this file.
   *
   * @var int
   */
  protected $notices;

  /**
   * Amount of warnings for this file.
   *
   * @var int
   */
  protected $warnings;

  /**
   * Amount of errors for this file.
   *
   * @var int
   */
  protected $errors;

  /**
   * Return a suitable source code formatter from given filename.
   *
   * @param string $file_name
   *   Path/name of a file.
   *
   * @return string
   *   Source code formatter identifier.
   *
   * @see Pareview::getFormattersByExtension
   */
  protected static function extractFormatter($file_name) {
    $formats = Config::getFormattersByExtension();
    $ext = strtolower(pathinfo($file_name, PATHINFO_EXTENSION));
    if (in_array($ext, $formats)) {
      return $formats[$ext];
    }
    else {
      return '';
    }
  }

  /**
   * Set the formatter for the file.
   *
   * This formatter is intended to being used in conjunction with
   * syntax highlighters (e.g. GeSHi) when showing source code.
   * Therefore changing the formatter here will change the formatter
   * of all issues' source code snippets (proofs) to the given
   * formatter.
   *
   * @param string $formatter
   *   Format string.
   *
   * @return \pareviewsh\IssuesInterface
   *   File issues.
   *
   * @see self::extractFormatter
   * @see Pareview::getFormattersByExtension
   */
  protected function setFormatter($formatter) {
    $this->formatter = $formatter;
    if (empty($this->issues)) {
      return;
    }
    foreach ($this->issues as $issue) {
      if (!empty($proof = $issue->getProof())) {
        $proof['format'] = $formatter;
        $issue->setProof($proof);
      }
    }

    return $this;
  }

  /**
   * Construct a file instance.
   *
   * @param string $file_name
   *   Path and name of the file.
   * @param bool $strip_base
   *   [Optional] Whether to strip the PAReview.sh workspace
   *   prefix from the file path (TRUE) or not (FALSE).
   *   Defaults to TRUE.
   */
  public function __construct($file_name, $strip_base = TRUE) {
    $this->issues = [];
    $this->formatter = '';
    $this->notices = 0;
    $this->warnings = 0;
    $this->errors = 0;
    $this->setFileName($file_name, $strip_base);
  }

  /**
   * Set path/name of the file.
   *
   * @param string $file_name
   *   Path and name of the file.
   * @param bool $strip_base
   *   [Optional] Whether to strip the PAReview.sh workspace
   *   prefix from the file path (TRUE) or not (FALSE).
   *   Defaults to TRUE.
   *
   * @return \pareviewsh\IssuesInterface
   *   File issues.
   */
  public function setFileName($file_name, $strip_base = TRUE) {
    // By default, we store file names relative to PAReview.sh
    // base dir. So remove it from file name, if the file name
    // starts with the base dir.
    $base = Pareview::getInstance()->getConfig()->workspace;
    if ($strip_base && strpos($file_name, $base) === 0) {
      $file_name = '.' . substr($file_name, strlen($base));
    }

    $this->fileName = $file_name;
    $formatter = self::extractFormatter($file_name);
    $this->setFormatter($formatter);

    return $this;
  }

  /**
   * Return path/name of the file.
   *
   * This path is relative to the workspace of PAReview.sh.
   *
   * @return string
   *   Path/name of the file.
   */
  public function getFileName() {
    return $this->fileName;
  }

  /**
   * {@inheritdoc}
   */
  public function hasIssues() {
    return count($this->issues) > 0;
  }

  /**
   * {@inheritdoc}
   */
  public function addIssue(IssueInterface $issue) {
    if (empty($issue->getFileName())) {
      $issue->setFileName($this->fileName);
    }

    if (!empty($proof = $issue->getProof())) {
      $proof['format'] = $this->formatter;
      $issue->setProof($proof);
    }

    $this->issues[] = $issue;

    switch ($issue->getSeverity()) {
      case IssueInterface::SEVERITY_NOTICE:
        $this->notices++;
        break;
      case IssueInterface::SEVERITY_WARNING:
        $this->warnings++;
        break;
      case IssueInterface::SEVERITY_ERROR:
        $this->errors++;
        break;
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getIssues() {
    return $this->issues;
  }

  /**
   * {@inheritdoc}
   */
  public function getNoticesCount() {
    return $this->notices;
  }

  /**
   * {@inheritdoc}
   */
  public function getWarningsCount() {
    return $this->warnings;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrorsCount() {
    return $this->errors;
  }

  /**
   * {@inheritdoc}
   */
  public function sortByOccurrence() {
    if (count($this->issues) > 1) {
      usort($this->issues, function ($item1, $item2) {
        if ($item1->getLine() == $item2->getLine()) {
          if ($item1->getColumn() == $item2->getColumn()) {
            return 0;
          }
          else {
            return $item1->getColumn() < $item2->getColumn() ? -1 : 1;
          }
        }
        else {
          return $item1->getLine() < $item2->getLine() ? -1 : 1;
        }
      });
    }

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return [
      'file' => $this->fileName,
      'file_formatter' => $this->formatter,
      'notices' => $this->notices,
      'warnings' => $this->warnings,
      'errors' => $this->errors,
      'issues' => $this->issues,
    ];
  }

}
