<?php

namespace pareviewsh;

use \JsonSerializable;
use \pareviewsh\Exceptions\RuntimeException;

/**
 * Issue information.
 */
class Issue implements IssueInterface, JsonSerializable {

  /**
   * Severity of the issue.
   *
   * @var string
   */
  protected $severity;

  /**
   * Human-readable description of the issue.
   *
   * @var string
   */
  protected $message;

  /**
   * Array of help links providing further information about this issue.
   *
   * Each link is keyed by the link title, having the related URL as value.
   *
   * @var array
   */
  protected $helpLinks;

  /**
   * ID of the test that identified the issue.
   *
   * @var string
   */
  protected $testId;

  /**
   * ID of the rule that identified the issue.
   *
   * @var string
   */
  protected $ruleId;

  /**
   * {@inheritdoc}
   */
  public function __construct($issue = NULL) {
    // Set default values.
    $this->resetDefaults();

    // Whether we don't have initial values.
    if (empty($issue)) {
      return;
    }
    // Whether we have a standard object with initial values.
    if (is_object($issue)) {
      $issue = (array) $issue;
    }
    // Whether the initial values are not valid.
    if (!is_array($issue) || empty($issue)) {
      return;
    }

    // Set all initial values from argument.
    foreach($issue as $key => $value) {
      $this->{$key} = $value;
    }
  }

  /**
   * Return property value from JSON serialized property names.
   *
   * @param string $name
   *   Property name.
   *
   * @return mixed
   *   Value of the property, if it can be resolved.
   *
   * @throws \pareviewsh\Exceptions\RuntimeException
   *   If the property name could not be resolved.
   */
  public function __get($name) {
    $check = $this->getPropertyName($name);
    if (!property_exists($this, $check)) {
      throw new RuntimeException("Unable to get value of property \"$name\".");
    }
    return $this->{$check};
  }

  /**
   * Set property value using JSON serialized property names.
   *
   * @param string $name
   *   Property name.
   * @param mixed $value
   *   Property value.
   *
   * @throws \pareviewsh\Exceptions\RuntimeException
   *   If the property name could not be resolved.
   */
  public function __set($name, $value) {
    $check = $this->getPropertyName($name);
    if (!property_exists($this, $check)) {
      throw new RuntimeException("Unable to set value of property \"$name\".");
    }
    $this->{$check} = $value;
  }

  /**
   * {@inheritdoc}
   */
  public function setSeverity($severity) {
    $this->severity = $severity;
  }

  /**
   * {@inheritdoc}
   */
  public function getSeverity() {
    return $this->severity;
  }

  /**
   * {@inheritdoc}
   */
  public function setMessage($message) {
    $this->message = $message;
  }

  /**
   * {@inheritdoc}
   */
  public function getMessage() {
    return $this->message;
  }

  /**
   * {@inheritdoc}
   */
  public function setHelpLinks(array $help_links) {
    $this->helpLinks = $help_links;
  }

  /**
   * {@inheritdoc}
   */
  public function getHelpLinks() {
    return $this->helpLinks;
  }


  /**
   * {@inheritdoc}
   */
  public function setTestId($test_id) {
    $this->testId = $test_id;
  }

  /**
   * {@inheritdoc}
   */
  public function getTestId() {
    return $this->testId;
  }

  /**
   * {@inheritdoc}
   */
  public function setRuleId($rule_id) {
    $this->ruleId = $rule_id;
  }

  /**
   * {@inheritdoc}
   */
  public function getRuleId() {
    return $this->ruleId;
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return [
      'severity' => $this->severity,
      'message' => $this->message,
      'help_links' => $this->helpLinks,
      'test_id' => $this->testId,
      'rule_id' => $this->ruleId,
    ];
  }

  /**
   * Convert lowercase underscore property names to camel case.
   *
   * @param string $name
   *   Lowercase underscore property name.
   * @param bool $firstCap
   *   [Optional] Whether to capitalize the first character.
   *   Defaults to FALSE.
   *
   * @return string
   *   Camel case property name.
   */
  protected function getPropertyName($name, $firstCap = FALSE) {
    $str = str_replace(' ', '', ucwords(str_replace('_', ' ', $name)));
    if (!$firstCap) {
      $str[0] = strtolower($str[0]);
    }
    return $str;
  }

  /**
   * Reset all properties to their initial values.
   */
  protected function resetDefaults() {
    $this->message = '';
    $this->helpLinks = [];
    $this->severity = self::SEVERITY_NONE;
    $this->testId = '';
    $this->ruleId = '';
  }

}
