<?php

namespace pareviewsh;

/**
 * Interface for the most basic information an issue provides.
 */
interface IssueInterface {

  /**
   * Empty issue severity.
   */
  const SEVERITY_NONE = '';

  /**
   * Issue severity for notices.
   */
  const SEVERITY_NOTICE = 'notice';

  /**
   * Issue severity for warnings.
   */
  const SEVERITY_WARNING = 'warning';

  /**
   * Issue severity for errors.
   */
  const SEVERITY_ERROR = 'error';

  /**
   * Construct an instance of an issue.
   *
   * @param object|array|null $issue
   *   [Optional] Initial issue values. Defaults to NULL and may be an
   *   empty value or a standard object/associative array with some or
   *   all of the issue properties.
   */
  public function __construct($issue = NULL);

  /**
   * Set severity of the issue.
   *
   * @param string $severity
   *   Issue severity. Allowed values are self::SEVERITY_NOTICE,
   *   self::SEVERITY_WARNING and self::SEVERITY_ERROR.
   */
  public function setSeverity($severity);

  /**
   * Return severity of the issue.
   *
   * @return string
   *   Issue severity. May be one of the values self::SEVERITY_NOTICE,
   *   self::SEVERITY_WARNING and self::SEVERITY_ERROR, or
   *   self::SEVERITY_NONE, if no severity has been set for this issue.
   */
  public function getSeverity();

  /**
   * Set human-readable description of the issue.
   *
   * @param string $message
   *   Description of the issue.
   */
  public function setMessage($message);

  /**
   * Return human-readable description of the issue.
   *
   * @param string $message
   *   Description of the issue.
   */
  public function getMessage();

  /**
   * Set help links providing further information about this issue.
   *
   * @param array $help_links
   *   Associative array with help links to set. Each link is a key-value
   *   pair keyed by the link title, having the related URL as value.
   */
  public function setHelpLinks(array $help_links);

  /**
   * Return help links providing further information about this issue.
   *
   * @return array
   *   Associative array with help links to set. Each link is a key-value
   *   pair keyed by the link title, having the related URL as value.
   */
  public function getHelpLinks();

  /**
   * Set ID of the test that identified the issue.
   *
   * @param string $test_id
   *   Test ID.
   */
  public function setTestId($test_id);

  /**
   * Return ID of the test that identified the issue.
   *
   * @return string
   *   Test ID.
   */
  public function getTestId();

  /**
   * Set ID of the rule that identified the issue.
   *
   * @param string $rule_id
   *   Rule ID.
   */
  public function setRuleId($rule_id);

  /**
   * Return ID of the rule that identified the issue.
   *
   * @return string
   *   Rule ID.
   */
  public function getRuleId();

}
