<?php

namespace pareviewsh;

use \JsonSerializable;
use \pareviewsh\Exceptions\CliException;
use \pareviewsh\Exceptions\RuntimeException as RuntimeException2;
use \pareviewsh\Utils\Output;
use \RuntimeException;

/**
 * Main wrapper for all PAReview.sh tests.
 *
 * This class provides means to alter test settings and gain
 * environment information.
 */
class Pareview implements IssueStatisticsInterface, JsonSerializable {

  /**
   * Current instance of PAReview.sh.
   *
   * The architecture of PAReview.sh assumes, there is only one
   * active instance of PAReview.sh. Therefore we bind this instance
   * as static property to the class and provide an according getter
   * for all components of PAReview.sh.
   *
   * @var Pareview
   */
  protected static $pareview;

  /**
   * Output handler.
   *
   * @var Output
   */
  protected static $output;

  /**
   * Run-time configuration of PAReview.sh.
   *
   * @var Config
   */
  protected $config;

  /**
   * Git repository information.
   *
   * @var GitRepository
   */
  protected $git;

  /**
   * Instance of PAReview.sh checks.
   *
   * @var Checks
   */
  protected $checks;

  /**
   * Name of the project that is reviewed.
   *
   * @var string
   */
  protected $projectName;

  /**
   * Identifier of the project that is reviewed.
   *
   * @var string
   */
  protected $projectIdentifier;

  /**
   * Version string of the project that is reviewed.
   *
   * @var string
   */
  protected $projectVersion;

  /**
   * Description of the project that is reviewed.
   *
   * @var string
   */
  protected $projectDescription;

  /**
   * Return the current instance of the PAReview.sh container.
   *
   * @return \pareviewsh\Pareview
   *   Current instance of PAReview.sh.
   *
   * @throws RuntimeException
   */
  public static function getInstance() {
    if (!isset(self::$pareview)) {
      throw new RuntimeException2('PAReview.sh is not initialized.', 3);
    }
    else {
      return self::$pareview;
    }
  }

  /**
   * Construct an instance of PAReview.sh.
   */
  public function __construct($is_cli = TRUE) {
    self::$pareview = $this;

    // Initialize project properties.
    $this->projectName = '';
    $this->projectDescription = '';
    $this->projectIdentifier = '';
    $this->projectVersion = '';

    try {
      self::$output = new Output(fopen('php://stdout', 'w'), $is_cli);
      $this->checks = new Checks();
      $this->config = new Config(self::$output, $is_cli);
    }
    catch (CliException $ex) {
      self::$output->writeLn($ex->getMessage());
      exit($ex->getCode());
    }
  }

  public function run() {
    if (!in_array($this->config->format, array_keys(Config::getFormats()))) {
      // ERROR!
    }

    // Run the checks.
    $exit_code = $this->checks->run();

    // Initialize output handler.
    $output = !empty($this->config->reportFile) ?
      new Output(fopen($this->config->reportFile, 'w'), $this->config->showColors)
      : self::$output;
    $output->enableFormatting($this->config->showColors);

    // Create the report.
    $report_class = "\\pareviewsh\\Reports\\" . str_replace('-', '', ucwords($this->config->format, '-')) . "Report";
    $report = new $report_class();
    $report->printReport($this, $output);

    return $exit_code;
  }

  /**
   * Return directory holding the source code to check.
   *
   * @return string
   */
  public function getWorkspaceDir() {
    return $this->config->workspace;
  }

  public function getChecks() {
    return $this->checks;
  }

  public function getConfig() {
    return $this->config;
  }

  public function setProjectName($project_name) {
    $this->projectName = $project_name;

    return $this;
  }

  public function getProjectName() {
    return $this->projectName;
  }

  public function setProjectIdentifier($project_identifier) {
    $this->projectIdentifier = $project_identifier;

    return $this;
  }

  public function getProjectIdentifier() {
    return $this->projectIdentifier;
  }

  public function setProjectVersion($project_version) {
    $this->projectVersion = $project_version;

    return $this;
  }

  public function getProjectVersion() {
    return $this->projectVersion;
  }

  /**
   * Set the description of the project investigated.
   *
   * @param string $project_description
   *   Project description.
   */
  public function setProjectDescription($project_description) {
    $this->projectDescription = $project_description;

    return $this;
  }

  /**
   * Return the description of the project investigated.
   *
   * @return string
   *   Project description.
   */
  public function getProjectDescription() {
    return $this->projectDescription;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrorsCount() {
    return $this->checks->getErrorsCount();
  }

  /**
   * {@inheritdoc}
   */
  public function getNoticesCount() {
    return $this->checks->getNoticesCount();
  }

  /**
   * {@inheritdoc}
   */
  public function getWarningsCount() {
    return $this->checks->getWarningsCount();
  }

  /**
   * {@inheritdoc}
   */
  public function hasIssues() {
    return $this->checks->hasIssues();
  }

  /**
   * {@inheritdoc}
   */
  public function jsonSerialize() {
    return [
      // General project information.
      'id' => $this->projectIdentifier,
      'name' => $this->projectName,
      'description' => $this->projectDescription,
      'version' => $this->projectVersion,
      'git' => $this->git,
      // Statistics.
      'has_issues' => $this->hasIssues(),
      'notices' => $this->getNoticesCount(),
      'warnings' => $this->getWarningsCount(),
      'errors' => $this->getErrorsCount(),
      // Check details.
      'checks' => $this->checks,
    ];
  }

}
