<?php

namespace pareviewsh\Parser;

use \pareviewsh\FileIssue;
use \pareviewsh\ParserInterface;
use \pareviewsh\Utils\SourceFile;

class CheckstyleXmlParser implements ParserInterface {

  /**
   * Prepares XML DOM nodes recursively for JSON encoding with attributes.
   *
   * @param \DOMNode $domNode
   *   XML DOM node.
   */
  protected static function prepareXml($domNode) {
    foreach ($domNode->childNodes as $node) {
      if ($node->hasChildNodes()) {
        self::prepareXml($node);
      }
      else {
        if ($domNode->hasAttributes() && strlen($domNode->nodeValue)) {
          $domNode->setAttribute("nodeValue", $node->textContent);
          $node->nodeValue = "";
        }
      }
    }
  }

  /**
   * Takes XML content and returns its data as array.
   *
   * @param string $content
   *   XML content.
   *
   * @return array
   *   Extracted XML data.
   */
  protected static function parseXml($content) {
    $dom = new \DOMDocument();
    $dom->loadXML($content);

    self::prepareXml($dom);
    $sxml = simplexml_load_string($dom->saveXML());

    $result = json_decode(json_encode($sxml), TRUE);
    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public static function parse($data, $testId = '', array $issues = []) {
    $data = self::parseXml($data);

    if (!empty($data['file'])) {
      foreach($data['file'] as $dataset) {
        $filename = $dataset['@attributes']['name'];
        $source_file = new SourceFile($filename);

        foreach (['error', 'warning'] as $severity) {
          if (!isset($dataset[$severity])) {
            continue;
          }
          if (isset($dataset[$severity]['@attributes'])) {
            $issue = [
              'severity' => $dataset[$severity]['@attributes']['severity'],
              'message' => $dataset[$severity]['@attributes']['message'],
              'file' => $filename,
              'line' => $dataset[$severity]['@attributes']['line'],
              'test_id' => $testId,
              'rule_id' => $dataset[$severity]['@attributes']['source'],
              'column' => $dataset[$severity]['@attributes']['column'],
            ];
            $issue['proof'] = $source_file->extractSource($issue['line']);

            $issues[] = new FileIssue($issue);
          }
          else {
            foreach($dataset['error'] as $details) {
              $issue = [
                'severity' => $details['@attributes']['severity'],
                'message' => $details['@attributes']['message'],
                'file' => $filename,
                'line' => $details['@attributes']['line'],
                'test_id' => $testId,
                'rule_id' => $details['@attributes']['source'],
                'column' => $details['@attributes']['column'],
              ];
              $issue['proof'] = $source_file->extractSource($issue['line']);

              $issues[] = new FileIssue($issue);
            }
          }
        }
        // Explicitly remove open file handles.
        $source_file = null;
      }
    }
    return $issues;
  }

}
