<?php

namespace pareviewsh\Reports;

use \pareviewsh\Pareview;
use \pareviewsh\ReportInterface;
use \pareviewsh\Utils\Output;
use \pareviewsh\Utils\MiscUtil;

/**
 * "classic" report class for PAReview.sh results.
 *
 * The "classic" format is a minimal HTML report format using
 * only H[2-4], UL, OL, LI, P, BR, B, I and CODE tags.
 *
 * It derived from the PAReview.sh 7.x-1.x output format and is
 * suitable for a comment in the drupal.org Project Applications
 * issue queue.
 */
class ScreenReport implements ReportInterface {

  /**
   * Output handler for the report.
   *
   * @var \pareviewsh\Utils\Output
   */
  protected $output;

  /**
   *
   * @var type \pareviewsh\Config
   */
  protected $config;

  protected $reportWidth;

  /**
   * Return a formatted severity label.
   *
   * @param string $severity
   *   Severity label.
   *
   * @return string
   *   Formatted severity label of 9 characters length.
   */
  protected function getSeverityLabel($severity) {
    $formatter = $this->output->format();

    // Whether colors are enabled.
    if ($formatter->isEnabled()) {
      $severity = strtolower($severity);
      $formatted = str_pad(strtoupper($severity), 9, ' ', STR_PAD_BOTH);
      if ($severity === 'error') {
        $formatted = $formatter($formatted)->bg_red()->white()->bold();
      }
      else if ($severity === 'warning') {
        $formatted = $formatter($formatted)->yellow()->bold();
      }
      else if ($severity === 'notice') {
        $formatted = $formatter($formatted)->bgDarkGray()->white()->bold();
      }
      return (string) $formatted;
    }

    // No color output.
    return '[' . str_pad(strtoupper($severity), 7, ' ', STR_PAD_BOTH) . ']';
  }

  /**
   * Return a formatted help links message from the given help links.
   *
   * The message will be like ` See [link1], [...] and [linkN].` for an
   * unlimited number of links, or empty if no help link was given.
   *
   * @param array $help_links<string, string>
   *   Associative help links array keyed by the link labels with link
   *   URLs as values.
   *
   * @return string
   *   Formatted help links message for the user.
   */
  protected function getHelpLinksMessage(array $help_links) {
    $message = '';
    if (empty($help_links)) {
      return $message;
    }

    $help_links = array_map(function($key, $value) {
      return $key . ' (' . $value . ')';
    }, array_keys($help_links), array_values($help_links));
    $message .= ' See ' . MiscUtil::getNativeList($help_links) . '.';

    return $message;
  }

  protected function printProject(array $project) {
    $output = $this->output;
    $title = '';
    if (!empty($project['project_name'])) {
      $title = 'project "' . $project['project_name'] . '"';
    }
    else if (!empty($project['project_id'])) {
      $title = 'project "' . $project['project_id'] . '"';
    }
    else if (!empty($this->config->gitRepository)) {
      $title = 'repository "' . $this->config->gitRepository . '"';
    }
    else if (!empty($this->config->directory)) {
      $title = 'directory "' . $this->config->directory . '"';
    }

    $output->writeLn($output->format(MiscUtil::wrapLn('Review ' . (empty($title) ? 'results' : 'of ' . $title), 0, 2, $this->reportWidth))->light_blue()->bold());

    if (empty($project['has_issues'])) {
      $output->writeLn(MiscUtil::wrapLn('Congratulations! Your project has no issues.', 0, 2, $this->reportWidth));
    }
    else {
      $total = $project['notices'] + $project['warnings'] + $project['errors'];
      $message = empty($project['warnings']) && empty($project['errors']) ? "Your project has no serious issues. However, the review found {$project['notices']} suggestions:" : ("The review found {$total} issues (" . MiscUtil::getSeverityCountList($project['notices'], $project['warnings'], $project['errors']) . "):");
      $output->writeLn(MiscUtil::wrapLn($message, 0, 2, $this->reportWidth));

      foreach ($project['checks'] as $check) {
        if (!empty($check['has_issues'])) {
          $this->printCheck($check);
        }
      }
    }
  }

  protected function printCheck(array $check) {
    $output = $this->output;
    $output->writeLn('');
    $output->writeLn($output->format(MiscUtil::wrap(strtoupper($check['name']), 0, 2, $this->reportWidth))->lightBlue()->bold());
    $output->writeLn(MiscUtil::wrapLn($check['description'], 0, 2, $this->reportWidth));

    if (!empty($check['issues'])) {
      foreach ($check['issues'] as $issue) {
        $severity = $this->getSeverityLabel($issue['severity']);
        $message = $issue['message'] . $this->getHelpLinksMessage($issue['help_links']);
        $output->writeLn('  ' . $severity . ' ' . MiscUtil::wrap($message, 0, 2, $this->reportWidth, '            '));
        if (!empty($issue['files'])) {
          $divider = "  " . $output->format(str_repeat('-', $this->reportWidth - 4))->light_gray();
          $output->writeLn('');
          foreach($issue['files'] as $file) {
            $output->writeLn($output->format("  File: " . $output->format(MiscUtil::wrap($file['file'], 0, 2, $this->reportWidth, '        '))->light_green())->bold());
            if (!$this->config->showProofs) {
              $output->writeLn($divider);
            }
            $last = array_values(array_slice($file['issues'], -1))[0];
            $max_length = strlen($last['line']);
            $prefix = '   ' . str_repeat(' ', $max_length) . ' |         | ';
            $source_prefix = '   ' . str_repeat(' ', $max_length) . ' | ';
            foreach ($file['issues'] as $line) {
              if ($this->config->showProofs) {
                $output->writeLn($divider);
              }
              $severity = $this->getSeverityLabel($line['severity']);
              $output->writeLn('   ' . str_pad($line['line'], $max_length, ' ', STR_PAD_LEFT) . ' |' . $severity . '| ' . MiscUtil::wrap($line['message'], 0, 3, $this->reportWidth, $prefix));
              if ($this->config->showProofs && !empty($line['proof'])) {
                $output->writeLn($divider);
                $start = $line['proof']['start'];
                $proof = $line['proof']['lines'];
                for ($i = 0; $i < count($proof); $i++) {
                  $cur = str_pad($start + $i, $max_length, ' ', STR_PAD_LEFT);
                  $code = $output->format('   ' . $cur . ' | ' . MiscUtil::wrap($proof[$i], 0, 3, $this->reportWidth, $source_prefix))->light_gray();
                  if ($start + $i == $line['line']) {
                    $code = $output->format($code)->bold();
                  }
                  $output->write($code);
                }
                $output->writeLn('');
              }
            }
            $output->writeLn('');
          }
        }
      }
    }
  }

  protected function printGeneralIssues($check) {
    
  }

  protected function printFileIssues($check) {
    
  }

  protected function printFileIssue($file_issue) {
    
  }

  /**
   * {@inheritdoc}
   */
  public function printReport(Pareview $pareview, Output $output) {
    $this->output = $output;
    $this->config = $pareview->getConfig();
    $this->reportWidth = MiscUtil::getTerminalWidth();

    $details = json_decode(json_encode($pareview), TRUE);

    $this->printProject($details);

//    if (empty($details['warnings']) && empty($details['errors'])) {
//      echo 'Great work! No issues found.' . PHP_EOL . PHP_EOL;
//    }
//    else {
//      echo '<ul>' . PHP_EOL;
//      foreach ($details['checks'] as $check) {
//        if ($check['has_issues']) {
//          if (!empty($check['issues_general'])) {
//            foreach ($check['issues_general'] as $issue) {
//              echo sprintf('<li>%s</li>' . PHP_EOL, $issue['message']);
//            }
//          }
//          if ($check['issues_files']) {
//            echo '<li>';
//            echo $check['issues_description'] . PHP_EOL;
//            echo '<ul>' . PHP_EOL;
//            foreach($check['issues_files'] as $file) {
//              echo '<li>' . $file['file'] . ' ' . PHP_EOL;
//              echo '<pre>';
//              $last = array_values(array_slice($file['issues'], -1))[0];
//              $max_length = strlen($last['line']);
//              $prefix = str_repeat(' ', $max_length) . ' |         | ';
//              foreach ($file['issues'] as $line) {
//                echo ' ' . str_pad($line['line'], $max_length) . ' | ' . str_pad(strtoupper($line['severity']), 7) . ' | ' . Terminal::wrapLn($line['message'], 1, 1, 70, $prefix);
//              }
//              echo '</pre></li>' . PHP_EOL;
//            }
//            echo '</ul>' . PHP_EOL;
//            echo '</li>' . PHP_EOL;
//          }
//        }
//      }
//      echo '</ul>' . PHP_EOL;
//    }
  }

}
