<?php

namespace pareviewsh\Tools;

use \pareviewsh\Config;
use \pareviewsh\Pareview;
use \pareviewsh\ToolInterface;
use \pareviewsh\Utils\CliCommand;

/**
 * CLI wrapper for PHP CodeSniffer.
 */
class CodeSniffer implements ToolInterface, ExtensionInterface {

  const PHPCS_REPORT_FULL = 'full';
  const PHPCS_REPORT_CHECKSTYLE = 'checkstyle';
  const PHPCS_REPORT_JSON = 'json';

  /**
   * Path to the CLI wrapper of PHPCS.
   *
   * @var string
   */
  protected static $command;

  /**
   * Ruleset directory.
   *
   * @var string
   */
  private $rulesetDir;

  /**
   * Report format.
   *
   * @var string
   */
  private $reportFormatter;

  /**
   * Output generated by running PHP CS.
   *
   * @var string
   */
  private $output;

  /**
   * Error messages generated by PHP CS.
   *
   * @var string
   */
  private $errors;

  /**
   * List of extensions for tool run.
   *
   * @var array
   */
  private $extensions;

  /**
   * Return the path to the CLI wrapper of PHPCS.
   *
   * This method assumes PHPCS to be installed using Composer. While
   * Unix systems can execute PHPCS scripts directly, the Windows
   * version relies on the wrappers within Composer's `bin` directory.
   * As the latter can be customized, we try to determine the
   * executable path using our `pareview.config` file that will
   * be created during Composer update/install by our Composer
   * hooks.
   *
   * @return string
   *   PHPCS path.
   */
  protected static function getCommand() {
    if (!isset(self::$command)) {
      // As fallback, we use the scripts located within the PHPCS
      // package.
      $fallback = '';
      $vendor_dir = Config::getVendorDir();
      // Whether PHPCS 2 is installed. This should always be the case for now.
      if (file_exists($vendor_dir . '/squizlabs/php_codesniffer/scripts/phpcs')) {
        $fallback = realpath($vendor_dir . '/squizlabs/php_codesniffer/scripts/phpcs');
      }
      // Whether PHPCS 3 is installed.
      else if (file_exists($vendor_dir . '/squizlabs/php_codesniffer/bin/phpcs')) {
        $fallback = realpath($vendor_dir . '/squizlabs/php_codesniffer/bin/phpcs');
      }

      self::$command = Config::getConfigFileData('BIN_PHPCS', $fallback);
    }

    return self::$command;
  }

  /**
   * {@inheritdoc}
   */
  public static function canRun() {
    // Whether a potentially executable file was found.
    $command = self::getCommand();
    if (empty($command)) {
      return FALSE;
    }

    // Try executing the command.
    $cli = new CliCommand();
    $output = '';
    $exit_code = $cli->exec($command . ' --version', NULL, $output);
    return ($exit_code === 0 && !empty($output) && empty($cli->getErrors()));
  }

  /**
   * Construct a PHP CS tool instance.
   *
   * @param string $ruleset_dir
   *   Directory with inspection rules.
   * @param string $report_formatter
   *   [Optional] Formatter to use. By default, we use `checkstyle`, as it
   *   is the most widely implemented standard for most our tools. This
   *   eases getting unified result sets by using the same parser.
   */
  public function __construct($ruleset_dir, $report_formatter = self::PHPCS_REPORT_CHECKSTYLE) {
    $this->rulesetDir = $ruleset_dir;
    $this->reportFormatter = $report_formatter;
    $this->output = '';
    $this->wrapper = '';
    $this->extensions = array_keys(Config::getFormattersByExtension());
  }

  /**
   * {@inheritdoc}
   */
  public function setExtensions(array $extensions) {
    $this->extensions = $extensions;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrors() {
    return $this->errors;
  }

  /**
   * {@inheritdoc}
   */
  public function getOutput() {
    return $this->output;
  }

  /**
   * {@inheritdoc}
   */
  public function run(array $files = []) {
    if (empty($files)) {
      $files = [Pareview::getInstance()->getConfig()->workspace];
    }

    foreach ($files as &$file) {
      $file = escapeshellarg($file);
    }

    $command = [self::getCommand()];
    $command[] = '--standard=' . escapeshellarg(realpath($this->rulesetDir));
    $command[] = '--report=' . escapeshellarg($this->reportFormatter);
    if (!empty($this->extensions)) {
      $command[] = '--extensions=' . escapeshellarg(implode(',', $this->extensions));
    }

    $cli = new CliCommand();
    $output = '';
    $exit_code = $cli->exec(implode(' ', $command) . ' ' . implode(' ', $files), NULL, $output);

    $this->output = $output;
    $this->errors = $cli->getErrors();

    return $exit_code;
  }

}
