<?php

namespace pareviewsh\Tools;

use \pareviewsh\Config;
use \pareviewsh\Pareview;
use \pareviewsh\ToolInterface;
use \pareviewsh\Utils\CliCommand;

/**
 * CLI wrapper for ESLint.
 */
class EsLint implements ToolInterface, ExtensionInterface {

  /**
   * Configuration file to use.
   *
   * @var string
   */
  private $configFile;

  /**
   * Whether to ignore `.eslintrc` files or not.
   *
   * @var bool
   */
  private $ignoreEsLintRc;

  /**
   * Whether to ignore minified files or not.
   *
   * @var bool
   */
  private $ignoreMinified;

  /**
   * Report format.
   *
   * @var string
   */
  private $reportFormatter;

  /**
   * Output generated by running ESLint.
   *
   * @var string
   */
  private $output;

  /**
   * Error messages generated by ESLint.
   *
   * @var string
   */
  private $errors;

  /**
   * List of extensions for tool run.
   *
   * @var array
   */
  private $extensions;

  /**
   * Return the path to the CLI wrapper of ESLint.
   *
   * This method assumes ESLint to be installed using the PAReview.sh
   * Composer hooks. (Running `npm` within the PAReview.sh root
   * directory.)
   *
   * @return string
   *   ESLint path.
   */
  protected static function getCommand() {
    $command = realpath(Config::getPareviewRoot() . '/node_modules/.bin/eslint');

    // Whether ESLint was found.
    if (file_exists($command)) {
      return $command;
    }

    return '';
  }

  /**
   * {@inheritdoc}
   */
  public static function canRun() {
    // Whether a potentially executable file was found.
    $command = self::getCommand();
    if (empty($command)) {
      return FALSE;
    }

    // Try executing the command.
    $cli = new CliCommand();
    $output = '';
    $exit_code = $cli->exec($command . ' --version', NULL, $output);
    return ($exit_code === 0 && !empty($output) && empty($cli->getErrors()));
  }

  /**
   * Construct a PHP CS tool instance.
   *
   * @param string $config_file
   *   Configuration file to use. (Any .eslintrc within the
   *   workspace will be ignored.)
   * @param string $report_formatter
   *   [Optional] Formatter to use. By default, we use `checkstyle`, as it
   *   is the most widely implemented standard for most our tools. This
   *   eases getting unified result sets by using the same parser.
   */
  public function __construct($config_file, $report_formatter = 'checkstyle', $ignore_eslintrc = TRUE, $ignore_minified = TRUE) {
    $this->configFile = $config_file;
    $this->reportFormatter = $report_formatter;
    $this->ignoreEsLintRc = $ignore_eslintrc;
    $this->ignoreMinified = $ignore_minified;
    $this->output = '';
    $this->errors = '';
    $this->extensions = ['js'];
  }

  /**
   * Turns `.eslintrc` ignore mode on or off.
   *
   * @param bool $enable
   *   TRUE to ignore `.eslintrc` files, FALSE otherwise.
   */
  public function setIgnoreEsLintRc($enable) {
    $this->ignoreEsLintRc = $enable;
  }

  /**
   * {@inheritdoc}
   */
  public function setExtensions(array $extensions) {
    $this->extensions = $extensions;
  }

  /**
   * {@inheritdoc}
   */
  public function getErrors() {
    return $this->errors;
  }

  /**
   * {@inheritdoc}
   */
  public function getOutput() {
    return $this->output;
  }

  /**
   * {@inheritdoc}
   */
  public function run(array $files = []) {
    if (empty($files)) {
      $files = [Pareview::getInstance()->getConfig()->workspace];
    }

    foreach ($files as &$file) {
      $file = escapeshellarg($file);
    }

    $command = [self::getCommand()];
    $command[] = '--config';
    $command[] = escapeshellarg($this->configFile);
    if ($this->ignoreEsLintRc) {
      $command[] = '--no-eslintrc';
    }
    $command[] = '--format';
    $command[] = escapeshellarg($this->reportFormatter);
    if (!empty($this->extensions)) {
      $command[] = '--ext';
      $command[] = escapeshellarg('.' . implode(',.', $this->extensions));
      if (!empty($this->ignoreMinified)) {
        $command[] = '--ignore-pattern';
        $command[] = escapeshellarg('**/*.min.' . implode(',**/*.min.', $this->extensions));
      }
    }

    $cli = new CliCommand();
    $output = '';
    $exit_code = $cli->exec(implode(' ', $command) . ' ' . implode(' ', $files), NULL, $output);

    $this->output = $output;
    $this->errors = $cli->getErrors();
    return $exit_code;
  }

}
