<?php

namespace pareviewsh\Utils;

use \Symfony\Component\Process\PhpProcess;
use \Symfony\Component\Process\Process;
use \Symfony\Component\Process\ProcessUtils;

/**
 * Wrapper for running commands on the CLI.
 *
 * @todo Check, whether using Symfony process component is of better use.
 */
class CliCommand {

  /**
   * Timeout after which a process should be killed.
   * @var type 
   */
  private $timeout;

  /**
   * Errors produced while running a cli command.
   *
   * @var string
   */
  private $errors;

  /**
   * Checks, whether a shell command is available or not.
   *
   * @param string $command
   *   Command to check for.
   *
   * @return bool
   *   TRUE, if the command is available, FALSE otherwise.
   */
  public function check($command) {
    $is_windows = strtolower(substr(PHP_OS, 0, 3)) === 'win';
    $check = ($is_windows) ? 'where' : 'command -v';

    $output = $errors = '';
    $return = $this->exec("{$check} {$command}", NULL, $output);

    return ($return !== FALSE) && empty($return) && empty($this->errors) && !empty($output);
  }

  /**
   * Kill the process with the given PID.
   *
   * @param int $pid
   *   Process ID of the system process to kill.
   *
   * @return int
   *   Return value of the kill command.
   */
  public function kill($pid){ 
    return (DIRECTORY_SEPARATOR === '\\') ? exec("taskkill /F /T /PID $pid") : exec("kill -9 $pid");
  }

  /**
   * Executes a shell command.
   *
   * @param string $command
   *   Command to check for.
   * @param string $cwd
   *   (Optional) The initial working dir for the command. This must be an
   *   absolute directory path, or NULL if you want to use the default value
   *   (the working dir of the current PHP process).
   * @param string $output
   *   (Optional) Variable to receive the STDOUT content produced by the
   *   command.
   * @param int $timeout
   *   [Optional] Timeout in seconds, after which the process should be
   *   killed.
   *   Defaults to 300.
   *
   * @return int|bool
   *   The return code of the command or FALSE, if it could not be executed.
   */
  public function exec($command, $cwd = NULL, &$output = '', $timeout = 300, $environment = NULL, $options = []) {
    // Reset errors.
    $this->errors = '';

    $process = new Process($command, $cwd, $environment, NULL, $timeout, $options);
    $process->run();
    $output = $process->getOutput();
    $this->errors = $process->getErrorOutput();

    return $process->getExitCode();
  }

//  /**
//   * Executes a shell command.
//   *
//   * @param string $command
//   *   Command to check for.
//   * @param string $cwd
//   *   (Optional) The initial working dir for the command. This must be an
//   *   absolute directory path, or NULL if you want to use the default value
//   *   (the working dir of the current PHP process).
//   * @param string $output
//   *   (Optional) Variable to receive the STDOUT content produced by the
//   *   command.
//   * @param int $timeout
//   *   [Optional] Timeout in seconds, after which the process should be
//   *   killed.
//   *   Defaults to 300.
//   *
//   * @return int|bool
//   *   The return code of the command or FALSE, if it could not be executed.
//   */
//  public function exec($command, $cwd = NULL, &$output = '', $timeout = 300) {
//    // Reset errors.
//    $this->errors = $errors = '';
//
//    $pipes = [];
//    // $start_time = microtime(TRUE);
//    echo $command . PHP_EOL;
//    $process = proc_open($command, [
//      0 => ['pipe', 'r'],
//      1 => ['pipe', 'w'],
//      2 => ['pipe', 'w'],
//        ], $pipes, $cwd, NULL, [
//      'bypass_shell' => FALSE,
//    ]);
//
//    if (is_resource($process)) {
//      $read_output = $read_error = TRUE;
//      $buffer_len = $prev_buffer_len = 0;
//      $ms = 10;
//      stream_set_blocking($pipes[1], 0);
//      stream_set_blocking($pipes[2], 0);
//
//      while ($read_error !== FALSE && $read_output !== FALSE) {
//        if ($read_output) {
//          if (feof($pipes[1])) {
//            fclose($pipes[1]);
//            $read_output = FALSE;
//          }
//          else {
//            $str = fgets($pipes[1], 1024);
//            $len = strlen($str);
//            if ($len) {
//              $output .= $str;
//              $buffer_len += $len;
//            }
//          }
//        }
//
//        if ($read_error) {
//          if (feof($pipes[2])) {
//            fclose($pipes[2]);
//            $read_error = FALSE;
//          }
//          else {
//            $str = fgets($pipes[2], 1024);
//            $len = strlen($str);
//            if ($len) {
//              $errors .= $str;
//              $buffer_len += $len;
//            }
//          }
//        }
//
//        if ($buffer_len > $prev_buffer_len) {
//          $prev_buffer_len = $buffer_len;
//          $ms = 10;
//        }
//        else {
//          usleep($ms * 1000); // sleep for $ms milliseconds
//          if ($ms < 160) {
//            $ms = $ms * 2;
//          }
//        }
//
////        // Whether the process is running too long.
////        $status = proc_get_status($process);
////        if (
////          ($read_error !== FALSE || $read_output !== FALSE)
////          && ($status !== FALSE)
////          && ($status['running'] === TRUE)
////          && (intval(microtime(TRUE) - $start_time) > $timeout)
////        ) {
////          self::kill($status['pid']);
////          echo "killed." . PHP_EOL;
////        }
//      }
//
//      $return = proc_close($process);
//
//      echo 'Output: ' . $output . PHP_EOL;
//      echo 'Errors: ' . $errors . PHP_EOL;
//      $this->errors = $errors;
//
//      return $return;
//    }
//
//    return FALSE;
//  }

  public function getErrors() {
    return $this->errors;
  }

  public function execPhp($command, $cwd = NULL, &$output = '', $timeout = 300, $environment = NULL, $options = []) {
    // Reset errors.
    $this->errors = '';

    $process = new PhpProcess($command, $cwd, $environment, $timeout, $options);
    $process->run();
    $output = $process->getOutput();
    $this->errors = $process->getErrorOutput();

    return $process->getExitCode();
  }

}
