<?php

namespace pareviewsh\Utils;

/**
 * General helper methods for PAReview.sh.
 */
class MiscUtil {

  /**
   * Whether the terminal discovery ran.
   *
   * @var bool
   */
  private static $isTerminalInitialized;

  /**
   * Width of the terminal in columns (characters).
   *
   * @var int
   */
  private static $terminalWidth;

  /**
   * Height of the terminal in lines.
   *
   * @var int
   */
  private static $terminalHeight;

  /**
   * Return width of the current terminal window.
   *
   * @param int $default
   *   [Optional] Default width to use. Defaults to 78.
   *
   * @return int
   *   Width of the terminal window.
   */
  public static function getTerminalWidth($default = 78) {
    if (!isset(self::$isTerminalInitialized)) {
      self::initTerminal();
    }

    return self::$terminalWidth ?: $default;
  }

  /**
   * Return the height of the current terminal window.
   *
   * @param int $default
   *   [Optional] Default height to use. Defaults to 32.
   *
   * @return int
   *   Height of the terminal window.
   *
   * @see self::getWidth
   */
  public static function getTerminalHeight($default = 32) {
    if (!isset(self::$isTerminalInitialized)) {
      self::initTerminal();
    }

    return self::$terminalHeight ?: $default;
  }

  /**
   * Wrap text for printing with limited width.
   *
   * @param string $text
   *   Text to print.
   * @param int $left_margin
   *   [Optional] Margin on the left side in characters.
   *   Defaults to `0`.
   * @param int $right_margin
   *   [Optional] Margin on the right side in characters.
   *   Defaults to `0`.
   * @param int $width
   *   [Optional] Available width. If a value `< 1` is given,
   *   the method attempts to use current terminal width as
   *   returned by self::getTerminalWidth.
   *   Defaults to `0`.
   *
   * @return string
   *   Wrapped text.
   */
  public static function wrap($text, $left_margin = 0, $right_margin = 0, $width = 0, $hanging_prefix = '', $suffix = '') {
    if (empty($width) || $width < 1) {
      $width = self::getTerminalWidth();
    }

    $width = $width - abs($left_margin) - abs($right_margin) - strlen($hanging_prefix) - strlen($suffix);
    $margin = str_repeat(' ', $left_margin);
    return $margin . wordwrap($text, $width, $suffix . PHP_EOL . $margin . $hanging_prefix, TRUE);
  }

  public static function wrapCommand($command, $left_margin = 0, $right_margin = 0, $width = 0, $hanging_prefix = '  ') {
    return self::wrap($command, $left_margin, $right_margin, $width, $hanging_prefix, self::getCommandSplit());
  }

  public static function wrapLn($text, $left_margin = 0, $right_margin = 0, $width = 0, $hanging_prefix = '', $suffix = '') {
    return self::wrap($text, $left_margin, $right_margin, $width, $hanging_prefix, $suffix) . PHP_EOL;
  }

  /**
   * Join the elements of an array to a native list.
   *
   * The native list will have all but the last two elements
   * joined with ', '. The last elements will be joined using
   * the given last concat word (e.g. 'and'/'or').
   *
   * @param array $values
   *   Values to join.
   * @param string $last_concat
   *   [Optional] Last concat word.
   *   Defaults to 'and'.
   *
   * @return string
   *   Native list with joined values.
   */
  public static function getNativeList(array $values, $last_concat = 'and') {
    return (implode(" {$last_concat} ", array_filter(array_merge(array(implode(', ', array_slice($values, 0, -1))), array_slice($values, -1)), 'strlen')));
  }

  /**
   * Return a human-readable summary of issues.
   *
   * @param int $suggestion
   *   Count of notices.
   * @param int $warning
   *   Count of warnings.
   * @param int $error
   *   Count of errors.
   *
   * @return string
   *   Summary of issues as native list.
   */
  public static function getSeverityCountList($suggestion, $warning, $error) {
    $issues = [];
    foreach (['error', 'warning', 'suggestion'] as $severity) {
      $value = $$severity;
      if (!empty($value)) {
        if ($value == 1) {
          $issues[] = $value . ' ' . $severity;
        }
        else {
          $issues[] = $value . ' ' . $severity . 's';
        }
      }
    }
    return empty($issues) ? 'no issues' : self::getNativeList($issues);
  }

  /**
   * A UTF8 compatible string pad
   *
   * @param string $text
   * @param int    $width
   * @param string $pad
   * @param int    $mode
   *
   * @return string
   */
  public static function pad($text, $width, $pad = ' ', $mode = STR_PAD_RIGHT) {
    $width = strlen($text) - mb_strlen($text, 'UTF-8') + $width;
    return str_pad($text, $width, $pad, $mode);
  }

  public static function getCommandSplit() {
    return strtolower(substr(PHP_OS, 0, 3)) === 'win' ? ' ^' : ' \\';
  }

  /**
   * Attempts to initialize terminal dimension information.
   *
   * This method first tries using STTY to gain terminal properties
   * information. This ensures the terminal will be correctlt determined
   * when running within any Linux distribution or Linux terminal for
   * Windows.
   *
   * If STTY is not available and Windows identified as host system, then
   * AnsiCon and Mode will be tested.
   */
  private static function initTerminal() {
    // Whether we have STTY on Unix systems, CygWin or MingW.
    if ($sttyString = self::getSttyColumns()) {
      if ($width = getenv('COLUMNS') && $height = getenv('LINES')) {
        self::$terminalWidth = (int) trim($width);
        self::$terminalHeight = (int) trim($height);
      }
      else if (preg_match('/rows.(\d+);.columns.(\d+);/i', $sttyString, $matches)) {
        // extract [w, h] from "rows h; columns w;"
        self::$terminalWidth = (int) $matches[2];
        self::$terminalHeight = (int) $matches[1];
      }
      else if (preg_match('/;.(\d+).rows;.(\d+).columns/i', $sttyString, $matches)) {
        // extract [w, h] from "; h rows; w columns"
        self::$terminalWidth = (int) $matches[2];
        self::$terminalHeight = (int) $matches[1];
      }
    }
    // Whether we are on Windows.and can use AnsiCon.
    else if (DIRECTORY_SEPARATOR === '\\' && preg_match('/^(\d+)x(\d+)(?: \((\d+)x(\d+)\))?$/', trim(getenv('ANSICON')), $matches)) {
      if ($width = getenv('COLUMNS') && $height = getenv('LINES')) {
        self::$terminalWidth = (int) trim($width);
        self::$terminalHeight = (int) trim($height);
      }
      // Extract width and height from AnsiCon information matches.
      else {
        self::$terminalWidth = (int) $matches[1];
        self::$terminalHeight = isset($matches[4]) ? (int) $matches[4] : (int) $matches[2];
      }
    }
    // Whether we are on Windows.and can use Mode command.
    else if (DIRECTORY_SEPARATOR === '\\' && $dimensions = self::getConsoleMode() !== NULL) {
      if ($width = getenv('COLUMNS') && $height = getenv('LINES')) {
        self::$terminalWidth = (int) trim($width);
        self::$terminalHeight = (int) trim($height);
      }
      // Extract width and height from Mode information.
      else {
        self::$terminalWidth = (int) $dimensions[0];
        self::$terminalHeight = (int) $dimensions[1];
      }
    }
    self::$isTerminalInitialized = TRUE;
  }

  /**
   * Runs and parses mode CON if it's available, suppressing any error output.
   *
   * @return int[]|null An array composed of the width and the height or null if it could not be parsed
   */
  private static function getConsoleMode() {
    if (!function_exists('proc_open')) {
      return;
    }

    $descriptorspec = array(
      1 => array('pipe', 'w'),
      2 => array('pipe', 'w'),
    );
    $process = proc_open('mode CON', $descriptorspec, $pipes, null, null, array('suppress_errors' => true));
    if (is_resource($process)) {
      $info = stream_get_contents($pipes[1]);
      fclose($pipes[1]);
      fclose($pipes[2]);
      proc_close($process);

      if (preg_match('/--------+\r?\n.+?(\d+)\r?\n.+?(\d+)\r?\n/', $info, $matches)) {
        return array((int) $matches[2], (int) $matches[1]);
      }
    }
  }

  /**
   * Runs and parses stty -a if it's available, suppressing any error output.
   *
   * @return string|null
   */
  private static function getSttyColumns() {
    if (!function_exists('proc_open')) {
      return;
    }

    $descriptorspec = array(
      1 => array('pipe', 'w'),
      2 => array('pipe', 'w'),
    );

    $process = proc_open('stty -a | grep columns', $descriptorspec, $pipes, null, null, array('suppress_errors' => true));
    if (is_resource($process)) {
      $info = stream_get_contents($pipes[1]);
      fclose($pipes[1]);
      fclose($pipes[2]);
      proc_close($process);

      return $info;
    }
  }

  /**
   * Sadly if you attempt to redirect stderr, e.g. "tput cols 2>/dev/null"
   * tput does not return the expected values. As a result, to prevent tput
   * from writing to stderr, we first check the exit code and call it again
   * to get the value :-(.
   *
   * @param int $default
   * @param string $param
   * @return int
   */
  private static function tput($default, $param = 'cols') {
    if ($param = 'cols' && isset(self::$terminalWidth)) {
      return self::$terminalWidth;
    } else if ($param = 'lines' && isset(self::$terminalHeight)) {
      return self::$terminalHeight;
    }

    // Whether we can run tput.
    $test = exec('tput ' . $param . ' 2>/dev/null');
    if (empty($test)) {
      return $default;
    }

    $result = intval(exec('tput ' . $param));
    if (!empty($result)) {
      if ($param = 'cols') {
        self::$terminalWidth = $result;
        return self::$terminalWidth;
      } else {
        self::$terminalHeight = $result;
        return self::$terminalHeight;
      }
    } else {
      return $default;
    }
  }

}
