<?php

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

namespace pareviewsh\Utils;

/**
 * Output wrapper for PAReview.sh.
 */
class Output {

  private $stream;

  /**
   * Output prettifier.
   *
   * @var \pareviewsh\Utils\Prettify
   */
  protected $prettify;

  /**
   * Construct an output instance.
   *
   * @param resource $stream
   *   A stream resource as `fopen('php://stdout', 'w')` for STDOUT or
   *   a file resource created like
   *   `fopen('/path/to/output.log', 'a', FALSE)`.
   * @param bool $decorated
   *   Whether to enable styles for the output.
   *
   * @throws \pareviewsh\Exceptions\RuntimeException
   *   When first argument is not a real stream.
   */
  public function __construct($stream, $decorated = FALSE) {
    if (!is_resource($stream) || 'stream' !== get_resource_type($stream)) {
      throw new \pareviewsh\Exceptions\RuntimeException('The Output class needs a stream as its first argument.');
    }

    $this->stream = $stream;
    $this->prettify = new \pareviewsh\Utils\Prettify();

    $this->enableFormatting($decorated);
  }

  /**
   * Gets the stream attached to this Output instance.
   *
   * @return resource
   *   A stream resource.
   */
  public function getStream() {
    return $this->stream;
  }

  /**
   * Return output formatter for the given text.
   *
   * @param string|null $text
   *   [Optional] Text to prettify.
   *   Defaults to NULL.
   *
   * @return \pareviewsh\Utils\Prettify
   *   Output formatter initialized with the given text.
   */
  public function format($text = NULL) {
    // Whether no text is given.
    if (is_null($text)) {
      return $this->prettify;
    }

    // Invoke callable, if text is given.
    $format = $this->prettify;
    return $format($text);
  }

  /**
   * Returns true if the stream supports colorization.
   *
   * Colorization is disabled if not supported by the stream:
   *
   *  -  Windows without Ansicon and ConEmu
   *  -  non tty consoles
   *
   * @return bool true if the stream supports colorization, false otherwise
   *
   * @link https://github.com/symfony/Console/blob/master/Output/StreamOutput.php#L95-L102
   * @codeCoverageIgnore
   */
  public function isFormattingSupported() {
    if (DIRECTORY_SEPARATOR === '\\') {
      return function_exists('sapi_windows_vt100_support') && sapi_windows_vt100_support($this->stream)
        || '10.0.10586' === PHP_WINDOWS_VERSION_MAJOR . '.' . PHP_WINDOWS_VERSION_MINOR . '.' . PHP_WINDOWS_VERSION_BUILD
        || false !== getenv('ANSICON')
        || 'ON' === getenv('ConEmuANSI')
        || 'xterm' === getenv('TERM');
    }

    return function_exists('posix_isatty') && @posix_isatty($this->stream);
  }

  public function enableFormatting($enabled) {
    $this->prettify->enableStyle($this->isFormattingSupported() && $enabled);
  }

  public function write($text) {
    $this->doWrite($text, FALSE);
  }

  public function writeLn($text) {
    $this->doWrite($text, TRUE);
  }

  /**
   * {@inheritdoc}
   */
  protected function doWrite($text, $newline) {
    if (@fwrite($this->stream, $text) === FALSE || ($newline && (@fwrite($this->stream, PHP_EOL) === FALSE))) {
      // should never happen
      throw new \pareviewsh\Exceptions\RuntimeException('Unable to write output.');
    }
    fflush($this->stream);
  }

}
