<?php

namespace pareviewsh\Utils;

/**
 * Helper to add colors and basic ANSI styles to text output.
 */
class Prettify {

  const FORMAT_PATTERN = '#<([a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*)>(.*?)</\\1?>#s';

  /** @link http://www.php.net/manual/en/functions.user-defined.php */
  const STYLE_NAME_PATTERN = '/^[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*$/';
  const ESC = "\033[";
  const ESC_SEQ_PATTERN = "\033[%sm";

  protected $initial = '';
  protected $wrapped = '';
  protected $userStyles = [];

  /**
   * Whether prettify is enabled or not.
   *
   * @var bool
   */
  protected $isEnabled = TRUE;
  // italic and blink may not work depending of your terminal
  protected $styles = [
    'reset' => '0',
    'bold' => '1',
    'dark' => '2',
    'italic' => '3',
    'underline' => '4',
    'blink' => '5',
    'reverse' => '7',
    'concealed' => '8',
    'default' => '39',
    'black' => '30',
    'red' => '31',
    'green' => '32',
    'yellow' => '33',
    'blue' => '34',
    'magenta' => '35',
    'cyan' => '36',
    'light_gray' => '37',
    'dark_gray' => '90',
    'light_red' => '91',
    'light_green' => '92',
    'light_yellow' => '93',
    'light_blue' => '94',
    'light_magenta' => '95',
    'light_cyan' => '96',
    'white' => '97',
    'bg_default' => '49',
    'bg_black' => '40',
    'bg_red' => '41',
    'bg_green' => '42',
    'bg_yellow' => '43',
    'bg_blue' => '44',
    'bg_magenta' => '45',
    'bg_cyan' => '46',
    'bg_light_gray' => '47',
    'bg_dark_gray' => '100',
    'bg_light_red' => '101',
    'bg_light_green' => '102',
    'bg_light_yellow' => '103',
    'bg_light_blue' => '104',
    'bg_light_magenta' => '105',
    'bg_light_cyan' => '106',
    'bg_white' => '107',
  ];

  /**
   * Create a prettify instance.
   *
   * @param string $text
   *   [Optional] Text to prettify.
   */
  public function __construct($text = '') {
    $this->setInternalState($text);
  }

  /**
   * Return this prettify instance initialized with the given text.
   *
   * @param string $text
   *   Text to prettify.
   *
   * @return \pareviewsh\Utils\Prettify
   *   Prettify instance to prettify the given text.
   */
  public function __invoke($text) {
    return $this->setInternalState($text);
  }

  /**
   * Match inaccessible method calls to styles.
   *
   * @param string $method
   *   Method name (style) called.
   * @param array $args
   *   [Optional] Array of arguments passed to the call.
   *
   * @return \pareviewsh\Utils\Prettify|string
   *   When called without text parameter, the method will
   *   apply the given style to the internal wrapped
   *   text and return the current instance for chaining.
   *   When called with text parameter, the styled text
   *   will be returned.
   */
  public function __call($method, $args) {
    if (count($args) >= 1) {
      return $this->apply($method, $args[0]);
    }

    return $this->apply($method);
  }

  public function __get($name) {
    return $this->apply($name);
  }

  public function __toString() {
    return $this->wrapped;
  }

  public function enableStyle($enabled) {
    $this->isEnabled = (bool) $enabled;
  }

  public function isEnabled() {
    return $this->isEnabled;
  }

  /**
   * Returns true if the stream supports colorization.
   *
   * Colorization is disabled if not supported by the stream:
   *
   *  -  Windows without Ansicon and ConEmu
   *  -  non tty consoles
   *
   * @return bool true if the stream supports colorization, false otherwise
   *
   * @link https://github.com/symfony/Console/blob/master/Output/StreamOutput.php#L95-L102
   * @codeCoverageIgnore
   */
  public function isSupported() {
    if (DIRECTORY_SEPARATOR === '\\') {
      return function_exists('sapi_windows_vt100_support') && sapi_windows_vt100_support($this->stream)
        || '10.0.10586' === PHP_WINDOWS_VERSION_MAJOR . '.' . PHP_WINDOWS_VERSION_MINOR . '.' . PHP_WINDOWS_VERSION_BUILD
        || false !== getenv('ANSICON')
        || 'ON' === getenv('ConEmuANSI')
        || 'xterm' === getenv('TERM');
    }

    return function_exists('posix_isatty') && @posix_isatty($this->stream);
  }

  /**
   * @codeCoverageIgnore
   */
  public function are256ColorsSupported() {
    return DIRECTORY_SEPARATOR === '/' && false !== strpos(getenv('TERM'), '256color');
  }

  /**
   * (Re-)initialize this instance with a given text.
   *
   * @param string $text
   *   New text to style.
   *
   * @return \pareviewsh\Utils\Prettify
   *   Instance of prettify initialized with the given text.
   */
  protected function setInternalState($text) {
    $this->initial = $this->wrapped = (string) $text;
    return $this;
  }

  /**
   * Apply a style to the given text.
   *
   * @param string $style
   *   Name of the style to apply. It can be either CamelCase, or
   *   lowercase/underscores and will be matched to the known styles.
   * @param string $text
   *   Text to prettify.
   *
   * @return string
   *   Text with applied styles, if styling is enabled and supported,
   *   the original text otherwise.
   *
   * @throws \pareviewsh\Exceptions\RuntimeException
   *   If styling is enabled, but the given style cannot be found.
   */
  protected function stylize($style, $text) {
    // Whether text prettification is enabled.
    if (!$this->shouldPrettify()) {
      return $text;
    }

    // Transform CamelCase styles to all lowercase divided by underscores.
    $style = strtolower(preg_replace('%([a-z])([A-Z])%', '\1_\2', $style));

    // Whether it is a user-defined style.
    if ($this->isUserStyleExists($style)) {
      return $this->applyUserStyle($style, $text);
    }

    // Whether it is a default style.
    if ($this->isStyleExists($style)) {
      return $this->applyStyle($style, $text);
    }

    // Whether it is a color.
    if (preg_match('/^((?:bg_)?)color\[([0-9]|[1-9][0-9]|1[0-9][0-9]|2[0-4][0-9]|25[0-5])\]$/', $style, $matches)) {
      $option = $matches[1] == 'bg_' ? 48 : 38;
      return $this->buildEscSeq("{$option};5;{$matches[2]}") . $text . $this->buildEscSeq($this->styles['reset']);
    }

    // No applicable style found.
    throw new \pareviewsh\Exceptions\RuntimeException("Invalid style $style");
  }

  /**
   * Whether to apply styles.
   *
   * @return bool
   *   TRUE, if styling is enabled and supported by the platform,
   *   FALSE otherwise.
   */
  protected function shouldPrettify() {
    return $this->isEnabled; // && $this->isSupported();
  }

  /**
   * Whether the given style exists.
   *
   * @param string $style
   *   Style name.
   *
   * @return bool
   *   TRUE, if the style exists, FALSE otherwise.
   */
  protected function isStyleExists($style) {
    return array_key_exists($style, $this->styles);
  }

  /**
   * Apply a style to text.
   *
   * @param string $style
   *   Style name.
   * @param string $text
   *   Text to style.
   *
   * @return string
   *   Text with applied styles.
   */
  protected function applyStyle($style, $text) {
    return $this->buildEscSeq($this->styles[$style]) . $text . $this->buildEscSeq($this->styles['reset']);
  }

  /**
   * Return an ESC sequence for the given style.
   *
   * @param string $style
   *   Style name.
   *
   * @return string
   *   ESC sequence for the style.
   */
  protected function buildEscSeq($style) {
    return sprintf(self::ESC_SEQ_PATTERN, $style);
  }

  /**
   * Whether a user-defined style exists.
   *
   * @param string $style
   *   Style name.
   *
   * @return bool
   *   TRUE, if a user-defined style with the given name exists,
   *   FALSE otherwise.
   */
  protected function isUserStyleExists($style) {
    return array_key_exists($style, $this->userStyles);
  }

  /**
   * Apply a user-defined style to the given text.
   *
   * @param string $user_style
   *   Style name.
   * @param string $text
   *   Text to be styled.
   *
   * @return string
   *   Text with applied user style.
   */
  protected function applyUserStyle($user_style, $text) {
    $styles = (array) $this->userStyles[$user_style];

    foreach ($styles as $style) {
      $text = $this->stylize($style, $text);
    }

    return $text;
  }

  /**
   * Apply a style to the given text and return the Prettify instance.
   *
   * Different from self::stylize, this method returns an instance of
   * the prettifier on order to allow chaining multiple styles.
   *
   * @param string $style
   *   Style name.
   * @param string $text
   *   [Optional] Text to style. Defaults to the wrapped text
   *   initialized during class creation or while calling this
   *   class as callable with text parameter.
   * @return $this
   */
  public function apply($style, $text = null) {
    if ($text === null) {
      $this->wrapped = $this->stylize($style, $this->wrapped);
      return $this;
    }

    return $this->stylize($style, $text);
  }

  public function fg($color, $text = null) {
    return $this->apply($color, $text);
  }

  public function bg($color, $text = null) {
    return $this->apply('bg_' . $color, $text);
  }

  public function highlight($color, $text = null) {
    return $this->bg($color, $text);
  }

  public function reset() {
    $this->wrapped = $this->initial;
    return $this;
  }

  public function center($width = 80, $text = null) {
    if ($text === null) {
      $text = $this->wrapped;
    }

    $centered = '';
    foreach (explode(PHP_EOL, $text) as $line) {
      $line = trim($line);
      $lineWidth = strlen($line) - mb_strlen($line, 'UTF-8') + $width;
      $centered .= str_pad($line, $lineWidth, ' ', STR_PAD_BOTH) . PHP_EOL;
    }

    $this->setInternalState(trim($centered, PHP_EOL));
    return $this;
  }

  protected function stripColors($text) {
    return preg_replace('/' . preg_quote(self::ESC) . '\d+m/', '', $text);
  }

  public function clean($text = null) {
    if ($text === null) {
      $this->wrapped = $this->stripColors($this->wrapped);
      return $this;
    }

    return $this->stripColors($text);
  }

  public function strip($text = null) {
    return $this->clean($text);
  }

  public function isAValidStyleName($name) {
    return preg_match(self::STYLE_NAME_PATTERN, $name);
  }

  /**
   * @deprecated
   * @codeCoverageIgnore
   */
  public function setTheme(array $theme) {
    return $this->setUserStyles($theme);
  }

  public function setUserStyles(array $userStyles) {
    foreach ($userStyles as $name => $styles) {
      if (!$this->isAValidStyleName($name)) {
        throw new InvalidStyleNameException("$name is not a valid style name");
      }
    }

    $this->userStyles = $userStyles;
    return $this;
  }

  protected function colorizeText($text) {
    return preg_replace_callback(self::FORMAT_PATTERN, array($this, 'replaceStyle'), $text);
  }

  /**
   * @link https://github.com/symfony/Console/blob/master/Formatter/OutputFormatter.php#L124-162
   */
  public function colorize($text = null) {
    if ($text === null) {
      $this->wrapped = $this->colorizeText($this->wrapped);
      return $this;
    }

    return $this->colorizeText($text);
  }

  protected function replaceStyle($matches) {
    return $this->apply($matches[1], $this->colorize($matches[2]));
  }

}
